import os
import filecmp
import re
import json
import shutil
from . import utils
from hamlet.loggers import logging
from hamlet.backend.common.fsutils import Search


logger = logging.getLogger("CONTEXT_TREE")


def format_unit_cf_dir(base_dir, level, unit, placement, region):
    # Determine placement by region if not explicitly provided
    if not placement:
        if region == "us-east-1":
            placement = "global"
        else:
            placement = "default"
    return os.path.join(base_dir, unit, placement)


def parse_stack_filename(filename):
    # Parse file name for key values
    # Assume Account is part of the stack filename
    basename = os.path.basename(filename)
    pattern = (
        r"([a-z0-9]+)-(.+)-([1-9][0-9]{10})-([a-z]{2}-[a-z]+-[1-9])(-pseudo)?-(.+)"
    )
    match = re.match(pattern, basename)
    if match:
        return dict(
            stack_level=match.group(1),
            stack_deployment_unit=match.group(2),
            stack_account=match.group(3),
            stack_region=match.group(4),
        )
    pattern = r"([a-z0-9]+)-(.+-.+)-(.+)-([a-z]{2}-[a-z]+-[1-9])(-pseudo)?-(.+)"
    match = re.match(pattern, basename)
    if match:
        return dict(
            stack_level=match.group(1),
            stack_deployment_unit=match.group(2),
            stack_region=match.group(4),
            stack_account="",
        )
    pattern = r"([a-z0-9]+)-(.+)-([a-z]{2}-[a-z]+-[1-9])(-pseudo)?-(.+)"
    match = re.match(pattern, basename)
    if match:
        return dict(
            stack_level=match.group(1),
            stack_deployment_unit=match.group(2),
            stack_region=match.group(3),
            stack_account="",
        )
    return dict(
        stack_level="", stack_deployment_unit="", stack_region="", stack_account=""
    )


def find_gen3_root_dir(dirname):

    if os.environ.get("ROOT_DIR"):
        return os.environ.get("ROOT_DIR")

    marked_root_dir = Search.upwards(dirname, "root.json")
    if marked_root_dir is not None:
        return os.path.dirname(marked_root_dir)

    config_root_dir = os.path.dirname(Search.upwards(dirname, "config"))
    infrastructure_root_dir = os.path.dirname(Search.upwards(dirname, "infrastructure"))
    root_dir = config_root_dir or infrastructure_root_dir
    if Search.isdir(root_dir, "config") and Search.isdir(root_dir, "infrastructure"):
        return root_dir

    return None


def find_gen3_tenant_dir(root_dir, tenant):
    patterns = [
        os.path.join("**", tenant, "tenant.json"),
        os.path.join("**", tenant, "config", "tenant.json"),
        os.path.join("**", "tenant.json"),
    ]
    matches = Search.match_dirs(*patterns, root=root_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find tenant dir")
    return matches[0]


def find_gen3_account_dir(root_dir, account):
    patterns = [
        os.path.join("**", account, "account.json"),
        os.path.join("**", account, "config", "account.json"),
    ]
    matches = Search.match_dirs(*patterns, root=root_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find account dir")
    return matches[0]


def find_gen3_account_state_dir(root_dir, account):
    patterns = [
        os.path.join("**", account, "state"),
        os.path.join("**", "state", "**", account),
        os.path.join("**", "infrastructure", "**", account),
        os.path.join("**", account, "infrastructure"),
    ]
    matches = Search.match_dirs(*patterns, root=root_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find account infrastructure dir")
    return matches[0]


def find_gen3_product_dir(root_dir, product):
    patterns = [
        os.path.join("**", product, "product.json"),
        os.path.join("**", product, "config", "product.json"),
    ]
    matches = Search.match_dirs(*patterns, root=root_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find product dir")
    return matches[0]


def find_gen3_product_infrastructure_dir(root_dir, product):
    patterns = [
        os.path.join("**", "infrastructure", "**", product),
        os.path.join("**", product, "infrastructure"),
    ]
    matches = Search.match_dirs(*patterns, root=root_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find product infrastructure dir")
    return matches[0]


def find_gen3_environment_dir(root_dir, product, environment):
    product_dir = find_gen3_product_dir(root_dir, product)
    patterns = [os.path.join("**", "solutionsv2", environment, "environment.json")]
    matches = Search.match_dirs(*patterns, root=product_dir, include_file_dirs=True)
    if not matches:
        raise Exception("Can't find environment dir")
    return matches[0]


def get_gen3_env(environment_obj, name, prefix):
    return environment_obj[prefix + name]


def set_gen3_env_dir(environment_object, env, prefix, *directories):
    for directory in directories:
        if os.path.isdir(directory):
            environment_object[prefix + env] = directory
            return True
    return False


def find_gen3_dirs(
    environment_obj,
    root_dir,
    tenant="",
    account="",
    product="",
    environment="",
    segment="",
    prefix="",
):
    e = environment_obj  # shortcut
    tenant = tenant or e.TENANT
    account = account or e.ACCOUNT
    product = product or e.PRODUCT
    environment = environment or e.ENVIRONMENT
    segment = segment or e.SEGMENT

    set_gen3_env_dir(e, "ROOT_DIR", prefix, root_dir)
    set_gen3_env_dir(e, "TENANT_DIR", prefix, find_gen3_tenant_dir(root_dir, tenant))
    set_gen3_env_dir(e, "ACCOUNT_DIR", prefix, find_gen3_account_dir(root_dir, account))
    set_gen3_env_dir(
        e,
        "ACCOUNT_STATE_DIR",
        prefix,
        find_gen3_account_state_dir(root_dir, account),
    )
    e[prefix + "ACCOUNT_SETTINGS_DIR"] = os.path.join(
        get_gen3_env(e, "ACCOUNT_DIR", prefix), "settings"
    )
    e[prefix + "ACCOUNT_OPERATIONS_DIR"] = os.path.join(
        get_gen3_env(e, "ACCOUNT_STATE_DIR", prefix), "operations"
    )

    if product:
        set_gen3_env_dir(
            e, "PRODUCT_DIR", prefix, find_gen3_product_dir(root_dir, product)
        )
        set_gen3_env_dir(
            e,
            "PRODUCT_INFRASTRUCTURE_DIR",
            prefix,
            find_gen3_product_infrastructure_dir(root_dir, product),
        )
        e[prefix + "PRODUCT_SETTINGS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_DIR", prefix), "settings"
        )
        e[prefix + "PRODUCT_SOLUTIONS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_DIR", prefix), "solutionsv2"
        )
        e[prefix + "PRODUCT_OPERATIONS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_INFRASTRUCTURE_DIR", prefix), "operations"
        )
        e[prefix + "PRODUCT_SHARED_SETTINGS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_SETTINGS_DIR", prefix), "shared"
        )
        e[prefix + "PRODUCT_SHARED_SOLUTIONS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_SOLUTIONS_DIR", prefix), "shared"
        )
        e[prefix + "PRODUCT_SHARED_OPERATIONS_DIR"] = os.path.join(
            get_gen3_env(e, "PRODUCT_OPERATIONS_DIR", prefix), "shared"
        )
        if environment:
            e[prefix + "ENVIRONMENT_SHARED_SETTINGS_DIR"] = os.path.join(
                get_gen3_env(e, "PRODUCT_SETTINGS_DIR", prefix), environment
            )
            e[prefix + "ENVIRONMENT_SHARED_SOLUTIONS_DIR"] = os.path.join(
                get_gen3_env(e, "PRODUCT_SOLUTIONS_DIR", prefix), environment
            )
            e[prefix + "ENVIRONMENT_SHARED_OPERATIONS_DIR"] = os.path.join(
                get_gen3_env(e, "PRODUCT_OPERATIONS_DIR", prefix), environment
            )
            if segment:
                e[prefix + "SEGMENT_SHARED_SETTINGS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_SETTINGS_DIR", prefix), "shared", segment
                )
                e[prefix + "SEGMENT_SHARED_SOLUTIONS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_SOLUTIONS_DIR", prefix), "shared", segment
                )
                e[prefix + "SEGMENT_SHARED_OPERATIONS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_OPERATIONS_DIR", prefix), "shared", segment
                )

                e[prefix + "SEGMENT_SETTINGS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_SETTINGS_DIR", prefix),
                    environment,
                    segment,
                )
                e[prefix + "SEGMENT_BUILDS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_SETTINGS_DIR", prefix),
                    environment,
                    segment,
                )
                e[prefix + "SEGMENT_SOLUTIONS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_SOLUTIONS_DIR", prefix),
                    environment,
                    segment,
                )
                e[prefix + "SEGMENT_OPERATIONS_DIR"] = os.path.join(
                    get_gen3_env(e, "PRODUCT_OPERATIONS_DIR", prefix),
                    environment,
                    segment,
                )


def upgrade_build_ref(root_dir, dry_run):
    legacy_files = Search.match_files(os.path.join("**", "build.ref"), root=root_dir)
    for legacy_file in legacy_files:
        logger.debug("Checking %s", legacy_file)
        upgraded_file = os.path.join(os.path.dirname(legacy_file), "build.json")
        if os.path.isfile(upgraded_file):
            continue
        logger.info("Upgrading %s", legacy_file)
        with open(legacy_file, "rt") as f:
            build_array = f.read().split(" ")
        try:
            build_data = {"Commit": build_array[0]}
        except IndexError as e:
            raise Exception(
                f"Unable to upgrade build reference in {legacy_file}"
            ) from e
        try:
            build_data["Tag"] = build_array[1]
        except IndexError:
            pass
        build_data["Formats"] = ["docker"]
        build_data = json.dumps(build_data, indent=4)
        if dry_run:
            logger.info(build_data)
            continue
        with open(upgraded_file, "wt+") as f:
            f.write(build_data)


def upgrade_shared_build_ref(root_dir, dry_run):
    legacy_files = Search.match_files(os.path.join("**", "*.ref"), root=root_dir)
    for legacy_file in legacy_files:
        if os.path.basename(legacy_file) == "build.ref":
            continue
        logger.debug("Checking %s", legacy_file)
        upgraded_file = os.path.join(os.path.dirname(legacy_file), "shared_build.json")
        if os.path.isfile(upgraded_file):
            continue
        logger.info("Upgrading %s", legacy_file)
        with open(legacy_file, "rt") as f:
            build_ref = f.read()
        upgraded_data = json.dumps({"Reference": build_ref}, indent=4)
        if dry_run:
            logger.info(upgraded_data)
            continue
        with open(upgraded_file, "wt+") as f:
            f.write(upgraded_data)


def upgrade_credentials(root_dir, dry_run):
    legacy_files = Search.match_files(
        os.path.join("**", "credentials.json"), root=root_dir
    )
    for legacy_file in legacy_files:
        logger.debug("Checking %s", legacy_file)
        with open(legacy_file, "rt") as f:
            legacy_data = json.load(f)
            credentials = legacy_data.get("Credentials")
            if credentials is None:
                return
        logger.info("Upgrading %s", legacy_file)
        upgraded_data = json.dumps(credentials or [], indent=4)
        if dry_run:
            logger.info(upgraded_data)
            continue
        with open(legacy_file, "wt+") as f:
            f.write(upgraded_data)


def upgrade_container_naming(root_dir, dry_run):
    legacy_files = Search.match_files(
        os.path.join("**", "container.json"), root=root_dir
    )
    for legacy_file in legacy_files:
        logger.debug("Checking %s", legacy_file)
        upgraded_file = os.path.join(os.path.dirname(legacy_file), "segment.json")
        if os.path.isfile(upgraded_file):
            continue
        logger.info("Upgrading %s", legacy_file)
        if dry_run:
            logger.info(f"{legacy_file}->{upgraded_file}")
            continue
        shutil.copy2(legacy_file, upgraded_file)


def upgrade_cmdb_repo_to_v1_0_0(root_dir, dry_run):
    # All build references now in json format
    upgrade_build_ref(root_dir, dry_run)
    # All shared build references now in json format
    upgrade_shared_build_ref(root_dir, dry_run)
    # Strip top level "Credentials" attribute from credentials
    upgrade_credentials(root_dir, dry_run)
    # Change of naming from "container" to "segment"
    upgrade_container_naming(root_dir, dry_run)
    return True


def cleanup_cmdb_repo_to_v1_0_0(root_dir, dry_run):
    # delete all .ref files. Because all of them converted into json format.
    legacy_files = Search.match_files(os.path.join("**", "*.ref"), root=root_dir)
    for legacy_file in legacy_files:
        logger.info("%sDeleting %s", dry_run, legacy_file)
        if dry_run:
            continue
        os.remove(legacy_file)
    # Change of naming from "container" to "segment"
    legacy_files = Search.match_files(
        os.path.join("**", "container.json"), root=root_dir
    )
    for legacy_file in legacy_files:
        logger.info("%sDeleting %s", dry_run, legacy_file)
        if dry_run:
            continue
        os.remove(legacy_file)
    return True


UPGRADE_V1_1_0_SOURCES = {
    "appsettings": "settings",
    "solutions": "solutionsv2",
    "credentials": "operations",
    "aws": "cf",
}


# Introduce separate environment and segment dirs
def upgrade_cmdb_repo_to_v1_1_0_settings(root_dir, dry_run, target_dir):
    # Create the shared file location for default segment
    shared_dir = os.path.join(target_dir, "shared")
    os.makedirs(shared_dir, exist_ok=True)
    sub_files = Search.list_files(root_dir)
    for sub_file in sub_files:
        src = os.path.join(root_dir, sub_file)
        dst = os.path.join(shared_dir, sub_file)
        logger.debug("Copying %s to %s", src, dst)
        if dry_run:
            continue
        shutil.copy2(src, dst)
    sub_dirs = Search.list_dirs(root_dir)
    for sub_dir in sub_dirs:
        environment = sub_dir
        sub_dir = os.path.join(root_dir, sub_dir)
        segment_dir = os.path.join(target_dir, environment, "default")
        os.makedirs(segment_dir, exist_ok=True)
        logger.debug("Copying %s to %s", sub_dir, segment_dir)
        if dry_run:
            continue
        for name in Search.list_all(sub_dir):
            src = os.path.join(sub_dir, name)
            dst = os.path.join(segment_dir, name)
            if os.path.isdir(src):
                shutil.copytree(src, dst)
            else:
                shutil.copy2(src, dst)

        # Remove anything unwanted
        segment_files = Search.match_files(
            os.path.join("**", "*.ref"),
            os.path.join("**", "container.json"),
            root=segment_dir,
        )
        for segment_file in segment_files:
            logger.debug("Deleting %s", segment_file)
            os.remove(segment_file)
    return True


def upgrade_cmdb_repo_to_v1_1_0_state(root_dir, dry_run, target_dir):
    # Create the shared file location
    shared_dir = os.path.join(target_dir, "shared")
    os.makedirs(shared_dir, exist_ok=True)
    # Copy across the shared files
    cf_dir = os.path.join(root_dir, "cf")
    if os.path.isdir(cf_dir):
        sub_files = Search.list_files(cf_dir)
        for sub_file in sub_files:
            src = os.path.join(cf_dir, sub_file)
            dst = os.path.join(shared_dir, sub_file)
            logger.debug("Copying %s to %s", src, dst)
            if dry_run:
                continue
            shutil.copy2(src, dst)
    # Process each sub dir
    sub_dirs = Search.list_dirs(root_dir)
    for sub_dir in sub_dirs:
        if sub_dir == "cf":
            continue
        environment = sub_dir
        segment_dir = os.path.join(target_dir, environment, "default")
        cf_dir = os.path.join(root_dir, sub_dir, "cf")
        if not os.path.isdir(cf_dir):
            continue
        os.makedirs(segment_dir, exist_ok=True)
        logger.debug("Copying %s to %s", cf_dir, segment_dir)
        if dry_run:
            continue
        for name in Search.list_all(cf_dir):
            src = os.path.join(cf_dir, name)
            dst = os.path.join(segment_dir, name)
            if os.path.isdir(src):
                shutil.copytree(src, dst)
            else:
                shutil.copy2(src, dst)


# config segment folders now under an environment folder
def upgrade_cmdb_repo_to_v1_1_0(root_dir, dry_run):
    for source in UPGRADE_V1_1_0_SOURCES:
        source_dirs = Search.match_dirs(os.path.join("**", source), root=root_dir)
        for source_dir in source_dirs:
            target_dir = os.path.join(
                os.path.dirname(source_dir), UPGRADE_V1_1_0_SOURCES[source]
            )
            logger.debug("Checking %s", source_dir)
            if os.path.isdir(target_dir):
                continue
            logger.info("Converting %s into %s", source_dir, target_dir)
            if source == "aws":
                upgrade_cmdb_repo_to_v1_1_0_state(source_dir, dry_run, target_dir)
            else:
                upgrade_cmdb_repo_to_v1_1_0_settings(source_dir, dry_run, target_dir)
            if dry_run:
                continue
            # Special processing
            if source == "solutions":
                # Shared solution files are specific to the default segment
                shared_default_dir = os.path.join(target_dir, "shared", "default")
                os.makedirs(shared_default_dir, exist_ok=True)
                target_shared_dir = os.path.join(target_dir, "shared")
                solution_files = Search.list_files(target_shared_dir)
                for solution_file in solution_files:
                    src = os.path.join(target_shared_dir, solution_file)
                    dst = os.path.join(shared_default_dir, solution_file)
                    logger.debug("Moving %s to %s", src, dst)
                    shutil.move(src, dst)
                # Process environments
                segment_files = Search.match_files(
                    os.path.join("**", "segment.json"), root=target_dir
                )
                for segment_file in segment_files:
                    segment_dir = os.path.dirname(segment_file)
                    environment_dir = os.path.dirname(segment_dir)
                    # Add environment.json file
                    with open(segment_file, "rt") as f:
                        segment = json.load(f)
                    environment_id = segment.get("Segment", {}).get("Environment")
                    environment_file = os.path.join(environment_dir, "environment.json")
                    logger.debug("Creating %s", environment_file)
                    with open(environment_file, "wt+") as f:
                        json.dump({"Environment": {"Id": environment_id}}, f)
                    logger.debug("Cleaning %s", segment_file)
                    segment_legacy_keys = ["Id", "Name", "Title", "Environment"]
                    for segment_legacy_key in segment_legacy_keys:
                        try:
                            del segment["Segment"][segment_legacy_key]
                        except KeyError:
                            pass
                    with open(segment_file, "wt") as f:
                        json.dump(segment, f)
                shared_segment_file = os.path.join(shared_default_dir, "segment.json")
                logger.debug("Creating %s", shared_segment_file)
                with open(shared_segment_file, "wt+") as f:
                    json.dump({"Segment": {"Id": "default"}}, f)
            elif source == "credentials":
                pem_files = Search.match_files(
                    os.path.join("**", "aws-ssh*.pem"), root=target_dir
                )
                for pem_file in pem_files:
                    filename = os.path.basename(pem_file)
                    segment_dir = os.path.dirname(pem_file)
                    # Move the pem files to make them invisible to the generation process
                    src = pem_file
                    dst = os.path.join(segment_dir, "." + filename)
                    logger.debug("Moving %s to %s", src, dst)
                    shutil.move(src, dst)
                    segment_ignore_file = os.path.join(segment_dir, ".gitignore")
                    if not os.path.isfile(segment_ignore_file):
                        logger.debug("Creaging %s", segment_ignore_file)
                        ignore_list = ["*.plaintext", "*.decrypted", "*.ppk"]
                        with open(segment_ignore_file, "wt+") as f:
                            f.write("\n".join(ignore_list))
    return True


def cleanup_cmdb_repo_to_v1_1_0(root_dir, dry_run):
    for source in UPGRADE_V1_1_0_SOURCES:
        source_dirs = Search.match_dirs(os.path.join("**", source), root=root_dir)
        for source_dir in source_dirs:
            target_dir = os.path.join(
                os.path.dirname(source_dir), UPGRADE_V1_1_0_SOURCES[source]
            )
            logger.debug("Checking %s", source_dir)
            logger.debug("Target dir %s", target_dir)
            if not os.path.isdir(target_dir):
                continue
            logger.info("%sDeleting %s", dry_run, source_dir)
            if dry_run:
                continue
            shutil.rmtree(source_dir)
    return True


def cleanup_cmdb_repo_to_v1_1_1(*args, **kwargs):
    # Rerun 1.1.0 to pick up errors in original implementation
    # Previously it only worked for product repos but now should
    # work for all repos
    return cleanup_cmdb_repo_to_v1_1_0(*args, **kwargs)


def cleanup_cmdb_repo_to_v2_0_0(root_dir, dry_run):
    config_dirs = Search.match_dirs(os.path.join("**", "config"), root=root_dir)
    for config_dir in config_dirs:
        solutions_dir = os.path.join(config_dir, "solutionsv2")
        if os.path.isdir(solutions_dir):
            logger.info("%sDeleting %s", dry_run, solutions_dir)
            if dry_run:
                continue
            shutil.rmtree(solutions_dir)
    return True


# container_* files now should be fragment_*
def upgrade_cmdb_repo_to_v1_2_0(root_dir, dry_run):
    legacy_files = Search.match_files(
        os.path.join("**", "container_*.ftl"), root=root_dir
    )
    for legacy_file in legacy_files:
        logger.debug("Checking %s", legacy_file)
        replacement_filename = os.path.basename(legacy_file).replace(
            "container_", "fragment_"
        )
        replacement_file = os.path.join(
            os.path.dirname(legacy_file), replacement_filename
        )
        if os.path.isfile(replacement_file):
            continue
        logger.info("%sRenaming %s to %s", dry_run, legacy_file, replacement_file)
        if dry_run:
            continue
        shutil.move(legacy_file, replacement_file)
    return True


def upgrade_cmdb_repo_to_v1_3_0(root_dir, dry_run):
    # Find accounts
    account_files = Search.match_files(
        os.path.join("**", "account.json"), root=root_dir
    )
    account_mappings = dict()
    for account_file in account_files:
        with open(account_file, "rt") as f:
            account = json.load(f)
            account_mappings[account["Account"]["AWSId"]] = account["Account"]["Id"]
    cf_dirs = Search.match_dirs(os.path.join("**", "cf"), root=root_dir)
    for cf_dir in cf_dirs:
        cmk_stacks = Search.match_files(
            os.path.join("**", "seg-cmk-*[0-9]-stack.json"), root=cf_dir
        )
        for cmk_stack in cmk_stacks:
            logger.info("Looking for CMK account in %s", cmk_stack)
            with open(cmk_stack, "rt") as f:
                cmk_stack_data = json.load(f)
            stack_outputs = cmk_stack_data["Stacks"][0]["Outputs"]
            cmk_account = None
            cmk_region = None
            for output in stack_outputs:
                if output["OutputKey"] == "Account":
                    cmk_account = output["OutputValue"]
                elif output["OutputKey"] == "Region":
                    cmk_region = output["OutputValue"]
            if cmk_account:
                cmk_account_id = account_mappings[cmk_account]
                cmk_path = os.path.dirname(cmk_stack)
                segment_stacks = Search.match_files(
                    os.path.join("**", "*stack.json"), root=cmk_path
                )
                for stack_file in segment_stacks:
                    parsed_stack = parse_stack_filename(stack_file)
                    stack_dir = os.path.dirname(stack_file)
                    stack_filename = os.path.basename(stack_file)
                    with open(stack_file, "rt") as f:
                        stack_data = json.load(f)
                    stack_outputs = stack_data["Stacks"][0]["Outputs"]
                    stackoutput_account = None
                    stackoutput_region = None
                    for output in stack_outputs:
                        if output["OutputKey"] == "Account":
                            stackoutput_account = output["OutputValue"]
                        elif output["OutputKey"] == "Region":
                            stackoutput_region = output["OutputValue"]
                    if not stackoutput_account:
                        logger.debug("Adding Account Output to %s", stack_file)
                        for stack in stack_data["Stacks"]:
                            stack["Outputs"].append(
                                {"OutputKey": "Account", "OutputValue": cmk_account}
                            )
                    if not stackoutput_region:
                        logger.debug("Adding Region Output to %s", stack_file)
                        for stack in stack_data["Stacks"]:
                            stack["Outputs"].append(
                                {
                                    "OutputKey": "Region",
                                    "OutputValue": parsed_stack["stack_region"],
                                }
                            )
                    if not stackoutput_region or not stackoutput_account:
                        with open(stack_file, "wt") as f:
                            json.dump(stack_data, f, indent=4)
                    if not parsed_stack["stack_account"]:
                        new_stack_file_name = os.path.basename(stack_file).replace(
                            f'-{parsed_stack["stack_region"]}-',
                            f'-{cmk_account_id}-{parsed_stack["stack_region"]}-',
                        )
                        if (
                            stack_filename != new_stack_file_name
                            and cmk_account_id not in stack_filename
                        ):
                            src = stack_file
                            dst = os.path.join(stack_dir, new_stack_file_name)
                            logger.debug("Moving %s to %s", src, dst)
                            if dry_run:
                                continue
                            shutil.move(src, dst)
                # Rename SSH keys to include Account/Region
                operations_path = cmk_path.replace(
                    os.path.join("infrastructure", "cf"),
                    os.path.join("infrastructure", "operations"),
                )
                logger.info("Checking for SSH Keys in %s", operations_path)
                pem_files = Search.match_files(
                    os.path.join("**", ".aws-ssh*.pem*"), root=operations_path
                )
                for pem_file in pem_files:
                    pem_dir = os.path.dirname(pem_file)
                    pem_basename = os.path.basename(pem_file)
                    new_basename = pem_basename.replace(
                        "aws-", f"aws-{cmk_account_id}-{cmk_region}-"
                    )
                    # Move the pem files to make them invisible to the generation process
                    src = pem_file
                    dst = os.path.join(pem_dir, new_basename)
                    logger.debug("Moving %s to %s", src, dst)
                    if dry_run:
                        continue
                    shutil.move(src, dst)
    return True


def upgrade_cmdb_repo_to_v1_3_1(root_dir, dry_run):
    # Find accounts
    account_files = Search.match_files(
        os.path.join("**", "account.json"), root=root_dir
    )
    account_mappings = dict()
    for account_file in account_files:
        with open(account_file, "rt") as f:
            account = json.load(f)
            account_mappings[account["Account"]["AWSId"]] = account["Account"]["Id"]
    cf_dirs = Search.match_dirs(os.path.join("**", "cf"), root=root_dir)
    for cf_dir in cf_dirs:
        cmk_stacks = Search.match_files(
            os.path.join("**", "seg-cmk-*[0-9]-stack.json"), root=cf_dir
        )
        for cmk_stack in cmk_stacks:
            logger.info("Looking for CMK account in %s", cmk_stack)
            with open(cmk_stack, "rt") as f:
                cmk_stack_data = json.load(f)
            stack_outputs = cmk_stack_data["Stacks"][0]["Outputs"]
            cmk_account = None
            for output in stack_outputs:
                if output["OutputKey"] == "Account":
                    cmk_account = output["OutputValue"]
            if cmk_account:
                cmk_account_id = account_mappings[cmk_account]
                cmk_path = os.path.dirname(cmk_stack)
                segment_cf = Search.match_files(os.path.join("**", "*"), root=cmk_path)
                for cf_file in segment_cf:
                    parsed_stack = parse_stack_filename(cf_file)
                    stack_dir = os.path.dirname(cf_file)
                    if not parsed_stack["stack_account"]:
                        cf_basename = os.path.basename(cf_file)
                        new_cf_basename = cf_basename.replace(
                            f'-{parsed_stack["stack_region"]}-',
                            f'-{cmk_account_id}-{parsed_stack["stack_region"]}-',
                        )
                        move_file = True
                        new_cf_file = os.path.join(stack_dir, new_cf_basename)
                        if (
                            cf_basename != new_cf_basename
                            and cmk_account_id not in cf_basename
                        ):
                            if os.path.isfile(new_cf_file):
                                if filecmp.cmp(cf_file, new_cf_file, False):
                                    move_file = False
                                else:
                                    logger.fatal(
                                        "Rename failed - %s already exists. Manual intervention necessary.",
                                        new_cf_file,
                                    )
                                    return False
                        if cf_file == new_cf_file:
                            logger.debug(
                                "Skipping %s, path is not changed", new_cf_file
                            )
                            continue
                        if move_file:
                            logger.debug("Moving %s to %s", cf_file, new_cf_file)
                        else:
                            logger.warn("%s already upgraded - removing", cf_file)
                        if dry_run:
                            continue
                        if move_file:
                            shutil.move(cf_file, new_cf_file)
                        else:
                            os.remove(cf_file)
    return True


def upgrade_cmdb_repo_to_v1_3_2(*args, **kwargs):
    # Rerun 1.3.1 to pick up errors in original implementation
    # Should be a no-op when run immediately after current 1.3.1
    # implementation
    return upgrade_cmdb_repo_to_v1_3_1(*args, **kwargs)


def upgrade_cmdb_repo_to_v2_0_0(root_dir, dry_run):
    # Reorganise cmdb to make it easier to manage via branches and dynamic cmdbs
    #
    # State is now in its own directory at the same level as config and infrastructure
    # Solutions is now under infrastructure
    # Builds are separated from settings and are now under infrastructure
    # Operations are now in their own directory at same level as config and
    # infrastructure. For consistency with config, a settings subdirectory has been
    # added.
    #
    # With this layout,
    # - infrastructure should be the same across environments assuming no builds
    #   are being promoted
    # - product and operations settings are managed consistently
    # - all the state info is cleanly separated (so potentially in its own repo)
    #
    # /config/settings
    # /operations/settings
    # /infrastructure/solutions
    # /infrastructure/builds
    # /state/cf
    # /state/cot
    #
    # If config and infrastructure are not in the one repo, then the upgrade must
    # be performed manually and the cmdb version manually updated
    config_dirs = Search.match_dirs(os.path.join("**", "config"), root=root_dir)
    for config_dir in config_dirs:
        base_dir = os.path.dirname(config_dir)
        solutions_dir = os.path.join(config_dir, "solutionsv2")
        settings_dir = os.path.join(config_dir, "settings")
        infrastructure_dir = os.path.join(base_dir, "infrastructure")
        state_dir = os.path.join(base_dir, "state")
        operations_dir = os.path.join(base_dir, "operations")
        state_subdirs = [
            os.path.join(infrastructure_dir, "cf"),
            os.path.join(infrastructure_dir, "cot"),
        ]
        if not os.path.isdir(infrastructure_dir):
            logger.warn(
                "%sUpdate to v2.0.0 for %s must be manually performed for split cmdb repos",
                dry_run,
                config_dir,
            )
            continue
        logger.debug("%sChecking %s", dry_run, base_dir)
        # Move the state into its own top level tree
        os.makedirs(base_dir, exist_ok=True)
        for state_subdir in state_subdirs:
            if os.path.isdir(state_subdir):
                src = state_subdir
                dst = os.path.join(state_dir, os.path.basename(state_subdir))
                logger.info("%sMoving %s to %s", dry_run, src, dst)
                if dry_run:
                    continue
                shutil.move(src, dst)
        # Move operations settings into their own top level tree
        orig_operations_settings_dir = os.path.join(infrastructure_dir, "operations")
        new_operation_settings_dir = os.path.join(operations_dir, "settings")
        if os.path.isdir(orig_operations_settings_dir):
            logger.info(
                "%sMoving %s to %s",
                dry_run,
                orig_operations_settings_dir,
                new_operation_settings_dir,
            )
            if dry_run:
                continue
            if not os.path.isdir(new_operation_settings_dir):
                os.makedirs(operations_dir, exist_ok=True)
                shutil.move(orig_operations_settings_dir, new_operation_settings_dir)
        # Copy the solutions tree from config to infrastructure and rename
        if os.path.isdir(solutions_dir):
            logger.info(
                "%sCopying %s to %s", dry_run, solutions_dir, infrastructure_dir
            )
            if not dry_run:
                # Leave existing solutions dir in place as it may be the current directory
                src = solutions_dir
                dst = os.path.join(infrastructure_dir, os.path.basename(src))
                shutil.copytree(src, dst)
            src = os.path.join(infrastructure_dir, "solutionsv2")
            dst = os.path.join(infrastructure_dir, "solutions")
            logger.info("%sRenaming %s to %s", dry_run, src, dst)
            if not dry_run:
                shutil.move(src, dst)
        # Copy the builds into their own tree
        builds_dir = os.path.join(infrastructure_dir, "builds")
        if not os.path.isdir(builds_dir):
            src = settings_dir
            dst = os.path.join(builds_dir, os.path.basename(src))
            logger.info("%sCopying %s to %s", dry_run, src, dst)
            if not dry_run:
                shutil.copytree(src, dst)
        # Remove the build files from the settings tree
        # Blob will pick up build references and shared builds
        logger.info("%sCleaning the settings tree", dry_run)
        setting_files = Search.match_files(
            os.path.join("**", "*build.json"), root=settings_dir
        )
        for setting_file in setting_files:
            logger.info("%sDeleting %s", dry_run, setting_file)
            if dry_run:
                continue
            os.remove(setting_file)
        # Build tree should only contain build references and shared builds
        logger.info("%sCleaning the builds tree", dry_run)
        if dry_run:
            build_files = Search.match_files(os.path.join("**", "*"), root=settings_dir)
        else:
            build_files = Search.match_files(os.path.join("**", "*"), root=builds_dir)
        build_files = [
            filename for filename in build_files if not filename.endswith("build.json")
        ]
        for build_file in build_files:
            logger.info("%sDeleting %s", dry_run, build_file)
            if dry_run:
                continue
            os.remove(build_file)

    return True


def upgrade_cmdb_repo_to_v2_0_1(root_dir, dry_run):
    # Reorganise state files into a directory tree based on deployment unit and placement
    #
    # The format of the state tree will follow the pattern
    # state/{df}/{env}/{seg}/{du}/{placement}
    #
    # Delete definition files because their file name contains the occurrence name not the
    # deployment unit. They will be regenerated into the correct dir on the next build.
    state_dirs = Search.match_dirs(os.path.join("**", "state"), root=root_dir)
    for state_dir in state_dirs:
        deployment_frameworks = Search.match_dirs("*", root=state_dir)
        for df_dir in deployment_frameworks:
            deployment_framework = os.path.basename(df_dir)
            logger.debug(
                "%sChecking %s deployment_framework", dry_run, deployment_framework
            )
            state_files = Search.match_files(os.path.join("**", "*"), root=df_dir)
            for state_file in state_files:
                state_basename = os.path.basename(state_file)
                state_dirname = os.path.dirname(state_file)
                stack_deployment_unit = ""
                # Filename format varies with deployment framework
                pattern_1 = r"([a-z0-9]+)-(.+)-([a-z][a-z0-9]+)-([a-z]{2}-[a-z]+-[1-9])(-pseudo)?-(.+)"
                pattern_2 = r"([a-z0-9]+)-(.+)-([a-z][a-z0-9]+)-(eastus|australiaeast|australiasoutheast|australiacentral|australiacentral2)(-pseudo)?-(.+)"  # noqa
                match = re.match(pattern_1, state_basename) or re.match(
                    pattern_2, state_basename
                )
                stack_level = ""
                stack_deployment_unit = ""
                stack_region = ""
                if match:
                    stack_level = match.group(1)
                    stack_deployment_unit = match.group(2)
                    stack_region = match.group(4)
                if not stack_deployment_unit:
                    # Legacy account formats
                    match = re.match(
                        r"account-([a-z][a-z0-9]+)-([a-z]{2}-[a-z]+-[1-9])-(.+)",
                        state_basename,
                    )
                    if match:
                        stack_level = "account"
                        stack_deployment_unit = match.group(1)
                        stack_region = match.group(2)
                    match = re.match(
                        r"account-([a-z]{2}-[a-z]+-[1-9])-(.+)", state_basename
                    )
                    if match:
                        stack_level = "account"
                        stack_deployment_unit = "s3"
                        stack_region = match.group(1)
                if not stack_deployment_unit:
                    # Legacy product formats
                    match = re.match(
                        r"product-([a-z]{2}-[a-z]+-[1-9])-(.+)", state_basename
                    )
                    if match:
                        stack_level = "product"
                        stack_deployment_unit = "cmk"
                        stack_region = match.group(1)
                if not stack_deployment_unit:
                    # Legacy segment formats
                    match = re.match(
                        r"seg-key-([a-z]{2}-[a-z]+-[1-9])-(.+)", state_basename
                    )
                    if match:
                        stack_level = "seg"
                        stack_deployment_unit = "cmk"
                        stack_region = match.group(1)
                    match = re.match(
                        r"cont-([a-z][a-z0-9]+)-([a-z]{2}-[a-z]+-[1-9])-(.+)",
                        state_basename,
                    )
                    if match:
                        stack_level = "seg"
                        stack_deployment_unit = match.group(1)
                        stack_region = match.group(2)
                if not stack_deployment_unit:
                    logger.warn(
                        "%sIgnoring %s, doesn't match one of the expected state filename formats",
                        dry_run,
                        state_basename,
                    )
                    continue
                if stack_level == "defn":
                    # Definition files are copied on every template creation
                    logger.info("%sDeleting %s", dry_run, state_file)
                    if dry_run:
                        continue
                    os.remove(state_file)
                else:
                    # Add deployment unit based subdirectories
                    if stack_deployment_unit in state_dirname:
                        logger.debug(
                            "%sIgnoring %s, already moved", dry_run, state_file
                        )
                    else:
                        du_dir = format_unit_cf_dir(
                            state_dirname,
                            stack_level,
                            stack_deployment_unit,
                            "",
                            stack_region,
                        )
                        src = state_file
                        dst = os.path.join(du_dir, state_basename)
                        logger.info("%sMoving %s to %s", dry_run, src, dst)
                        if dry_run:
                            continue
                        if not os.path.isdir(du_dir):
                            os.makedirs(du_dir, exist_ok=True)
                        shutil.move(src, dst)
    return True


GEN3_COMPATIBILITY = {}
# Entry example:  ["2.0.0"]=">=7.0.0"


def is_upgrade_compatible(cmdb_version, gen3_version):
    cmdb_version = utils.semver_clean(cmdb_version)
    compatible_range = GEN3_COMPATIBILITY.get(cmdb_version)
    if compatible_range:
        if not gen3_version:
            return "unknown"
        if not utils.semver_satisfies(gen3_version, compatible_range):
            return "incompatible"
    return "compatible"


def process_cmdb(root_dir, action, gen3_version, versions, dry_run):
    cmdb_git_repos = Search.match_dirs(os.path.join("**", "*.git"), root=root_dir)
    dry_run = "(Dryrun)" if dry_run else ""

    for cmdb_git_repo in cmdb_git_repos:

        cmdb_repo = os.path.dirname(cmdb_git_repo)
        cmdb_version_file = os.path.join(cmdb_repo, ".cmdb")
        current_version = ""
        pin_version = ""
        logger.debug("Checking repo %s", cmdb_repo)
        if os.path.isfile(cmdb_version_file):
            with open(cmdb_version_file, "rt") as f:
                cmdb_version_data = json.load(f)
            current_version = cmdb_version_data.get("Version", {}).get(
                action.capitalize()
            )
            pin_version = cmdb_version_data.get("Pin", {}).get(action.capitalize())
            logger.debug("Repo pinned at %s version %s", pin_version, current_version)
        else:
            with open(cmdb_version_file, "wt+") as f:
                json.dump({}, f)

        current_version = current_version or "v0.0.0"
        if utils.semver_compare(current_version, versions[-1]) >= 0:
            logger.debug(
                '%s of repo "%s" to %s is not required - skipping all version checks',
                action.capitalize(),
                cmdb_repo,
                versions[-1],
            )
            continue

        for version in versions:
            if utils.semver_compare(current_version, version) >= 0:
                logger.debug(
                    '%s of repo "%s" to %s is not required',
                    action.capitalize(),
                    cmdb_repo,
                    version,
                )
            else:
                logger.info(
                    '%s%s of repo "%s" to %s required ...',
                    dry_run,
                    action.capitalize(),
                    cmdb_repo,
                    version,
                )
            if pin_version:
                if utils.semver_compare(current_version, pin_version) < 0:
                    logger.warn(
                        '%s of repo "%s" to %s prevented by pin version %s',
                        action.capitalize(),
                        cmdb_repo,
                        version,
                        pin_version,
                    )
                    break
                else:
                    logger.debug(
                        '%s%s of repo "%s" to %s permitted by pin version %s',
                        dry_run,
                        action.capitalize(),
                    )
            compatibility = is_upgrade_compatible(version, gen3_version)
            if compatibility == "incompatible":
                logger.warn(
                    (
                        '%s%s of repo "%s" to %s is not compatible with the current gen3 framework version of %s. '
                        "Skipping upgrade process ..."
                    ),
                    dry_run,
                    action.capitalize(),
                    cmdb_repo,
                    version,
                    gen3_version,
                )
                break
            elif compatibility == "unknown":
                logger.warn(
                    (
                        '%s%s of repo "%" to %s requires the GEN3 framework version to be defined.'
                        "Skipping upgrade process ..."
                    ),
                    dry_run,
                    action.capitalize(),
                    cmdb_repo,
                    version,
                )
                break
            else:
                logger.debug(
                    '%s%s of repo "%s" to %s is compatible',
                    dry_run,
                    action.capitalize(),
                    cmdb_repo,
                    version,
                )

            cmdb_action_func = globals()[
                f'{action.lower()}_cmdb_repo_to_{version.replace(".", "_")}'
            ]
            if cmdb_action_func(cmdb_repo, dry_run):
                if dry_run:
                    logger.debug("%sSkipping later versions", dry_run)
                    break
                logger.info(
                    '%s of repo "%s" to %s successful',
                    action.capitalize(),
                    cmdb_repo,
                    version,
                )
                with open(cmdb_version_file, "rt") as f:
                    cmdb_version_data = json.load(f)
                utils.deep_dict_update(
                    cmdb_version_data, {"Version": {action.capitalize(): version}}
                )
                with open(cmdb_version_file, "wt") as f:
                    json.dump(cmdb_version_data, f)
                current_version = version

    return True


def upgrade_cmdb(root_dir, gen3_version, dry_run, maximum_version):
    maximum_version = maximum_version or "v1.3.2"
    upgrade_order = [
        "v1.0.0",
        "v1.1.0",
        "v1.2.0",
        "v1.3.0",
        "v1.3.1",
        "v1.3.2",
        "v2.0.0",
        "v2.0.1",
    ]
    logger.debug("Maximum CMDB upgrade version required is %s", maximum_version)
    required_versions = utils.semver_upgrade_list(upgrade_order, maximum_version)
    logger.debug("Required CMDB upgrade order is %s", required_versions)

    return process_cmdb(root_dir, "upgrade", gen3_version, required_versions, dry_run)


def cleanup_cmdb(root_dir, gen3_version, dry_run, maximum_version):
    maximum_version = maximum_version or "v1.1.1"
    upgrade_order = ["v1.0.0", "v1.1.0", "v1.1.1", "v2.0.0"]
    logger.debug("Maximum CMDB cleanup version required is %s", maximum_version)
    required_versions = utils.semver_upgrade_list(upgrade_order, maximum_version)
    logger.debug("Required CMDB cleanup order is %s", required_versions)

    return process_cmdb(root_dir, "cleanup", gen3_version, required_versions, dry_run)



from hamlet.backend.common import runner


def run(
    alias=None,
    base64_decode=None,
    decrypt=None,
    encrypt=None,
    crypto_file=None,
    key_id=None,
    no_alteration=None,
    json_path=None,
    quiet=None,
    crypto_text=None,
    update=None,
    visible=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    **kwargs,
):
    options = {
        "-a": alias,
        "-b": base64_decode,
        "-d": decrypt,
        "-e": encrypt,
        "-f": crypto_file,
        "-k": key_id,
        "-n": no_alteration,
        "-p": json_path,
        "-q": quiet,
        "-t": crypto_text,
        "-u": update,
        "-v": visible,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("manageCrypto.sh", [], options, env, _is_cli)



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.pipeline import pipeline as run_pipeline
from tests.unit.command.test_option_generation import run_options_test

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-i,--component"] = "component"
ALL_VALID_OPTIONS["!-t,--tier"] = "tier"
ALL_VALID_OPTIONS["-x,--instance"] = "instance"
ALL_VALID_OPTIONS["-y,--version"] = "version"
ALL_VALID_OPTIONS["-s,--pipeline-status-only"] = [True, False]
ALL_VALID_OPTIONS["-c,--pipeline-allow-concurrent"] = [True, False]


@mock.patch("hamlet.command.run.pipeline.run_pipeline_backend")
def test_input_valid(run_pipeline_backend):
    run_options_test(
        CliRunner(), run_pipeline, ALL_VALID_OPTIONS, run_pipeline_backend.run
    )



import os
import json
from marshmallow import ValidationError
from hamlet.backend.common.exceptions import BackendException
from .render import create_script
from .diagram_schema import Diagram as DiagramSchema


DIAGRAM_OUTPUT_PREFIX = "diagram"
DIAGRAM_CONFIG_OUTPUT_SUFFIX = "-config.json"
DIAGRAM_SCRIPT_OUTPUT_SUFFIX = "-script.py"
DIAGRAM_IMAGE_OUTPUT_SUFFIX = ".png"


def run(
    diagram_id,
    src_dir,
    output_dir,
):
    file_path = ""
    diagram_prefix = f"{DIAGRAM_OUTPUT_PREFIX}-{diagram_id}-"

    for name in os.listdir(src_dir):
        if name.startswith(diagram_prefix) and name.endswith(
            DIAGRAM_CONFIG_OUTPUT_SUFFIX
        ):
            file_path = os.path.join(src_dir, name)

    # if after all no files found raise an error
    if not file_path:
        raise BackendException("No diagram file found")

    diagram = dict()
    with open(file_path, "rt") as f:
        diagram_file_data = json.load(f)
        try:
            DiagramSchema().load(diagram_file_data)
        except ValidationError as e:
            message = json.dumps(e.messages, indent=4)
            raise BackendException(
                f"Invalid diagram schema in {file_path}\n\nErrors: \n{message}"
            ) from e
        diagram.update(**diagram_file_data)

    # if files have no diagrams data
    if not diagram:
        raise BackendException("No diagram found!")

    # Create outputs
    script_file_path = file_path.replace(
        DIAGRAM_CONFIG_OUTPUT_SUFFIX, DIAGRAM_SCRIPT_OUTPUT_SUFFIX
    )
    image_file_path = os.path.join(
        output_dir,
        os.path.basename(file_path).replace(
            DIAGRAM_CONFIG_OUTPUT_SUFFIX, DIAGRAM_IMAGE_OUTPUT_SUFFIX
        ),
    )

    text = create_script(
        diagram=diagram, temp_path=script_file_path, image_filename=image_file_path
    )
    if text is not None:
        with open(script_file_path, "w+t") as f:
            f.write(str(text))
    exec(text)
    return text



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.expo_app_publish import expo_app_publish as run_expo_app_publish
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "unit"
ALL_VALID_OPTIONS["-s,--run-setup"] = [True, False]
ALL_VALID_OPTIONS["-f,--force-binary-build"] = [True, False]
ALL_VALID_OPTIONS["-m,--submit-binary"] = [True, False]
ALL_VALID_OPTIONS["-b,--binary-build-process"] = "fastlane"
ALL_VALID_OPTIONS["-l,--build-logs"] = [True, False]
ALL_VALID_OPTIONS["-e,--environment-badge"] = [True, False]
ALL_VALID_OPTIONS["-d,--environment-badge-content"] = "environment"
ALL_VALID_OPTIONS["-n,--node-package-manager"] = "yarn"
ALL_VALID_OPTIONS["-v,--app-version-source"] = "cmdb"


@mock.patch("hamlet.command.run.expo_app_publish.run_expo_app_publish_backend")
def test_input_valid(run_expo_app_publish_backend):
    run_options_test(
        CliRunner(),
        run_expo_app_publish,
        ALL_VALID_OPTIONS,
        run_expo_app_publish_backend.run,
    )


@mock.patch("hamlet.command.run.expo_app_publish.run_expo_app_publish_backend")
def test_input_validation(run_expo_app_publish_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        run_expo_app_publish,
        run_expo_app_publish_backend.run,
        {
            "-u": "unit",
        },
        [],
    )



import os
import click


class GlobalEnvironmentConfig:
    """
    Handles Global Configuration which is used across hamlet
    """

    def __init__(self):
        self._home_dir = os.environ.get(
            "HAMLET_HOME_DIR",
            click.get_app_dir(app_name="hamlet", force_posix=True, roaming=False),
        )
        self._config_dir = os.environ.get(
            "HAMLET_CLI_CONFIG_DIR", os.path.join(self._home_dir, "config")
        )
        self._cli_cache_dir = os.environ.get(
            "HAMLET_CLI_CACHE_DIR", os.path.join(self._home_dir, "cli_cache")
        )

        self._engine = None
        self._engine_env = {}

    @property
    def home_dir(self):
        return self._home_dir

    @property
    def config_dir(self):
        return self._config_dir

    @property
    def cli_cache_dir(self):
        return self._cli_cache_dir

    @property
    def engine(self):
        return self._engine

    @engine.setter
    def engine(self, engine_name):
        self._engine = engine_name

    @property
    def engine_environment(self):
        return self._engine_env

    @engine_environment.setter
    def engine_environment(self, environment):
        self._engine_env = environment


HAMLET_GLOBAL_CONFIG = GlobalEnvironmentConfig()



from hamlet.backend.common import runner


def run(
    log_level=None,
    deployment_unit=None,
    code_commit=None,
    code_tag=None,
    image_format=None,
    registry_scope=None,
    _is_cli=False,
    env={},
):
    env = {
        "AUTOMATION_LOG_LEVEL": log_level,
        "DEPLOYMENT_UNITS": deployment_unit,
        "CODE_COMMIT": code_commit,
        "CODE_TAG": code_tag,
        "IMAGE_FORMATS": image_format,
        "REGISTRY_SCOPE": registry_scope,
        **env,
    }
    runner.run(
        "updateBuildReferences.sh",
        [],
        {},
        env,
        _is_cli,
        script_base_path_env="AUTOMATION_DIR",
    )



import collections
import tempfile
import hashlib
import json
import os

from unittest import mock
from click.testing import CliRunner
from hamlet.command.schema import create_schemas as create_schemas
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-o,--output-dir"] = "output_dir"
ALL_VALID_OPTIONS["-s,--schema"] = "Schema1"


def template_backend_run_mock(data):
    def run(
        entrance="schemalist",
        output_filename="schemalist-schemacontract.json",
        deployment_mode=None,
        output_dir=None,
        generation_input_source=None,
        generation_provider=None,
        generation_framework=None,
        log_level=None,
        root_dir=None,
        tenant=None,
        account=None,
        product=None,
        environment=None,
        segment=None,
        entrance_parameters=None,
    ):
        os.makedirs(output_dir, exist_ok=True)
        unitlist_filename = os.path.join(output_dir, output_filename)
        with open(unitlist_filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(schemalist=None):
    def decorator(func):
        @mock.patch("hamlet.command.schema.create_template_backend")
        @mock.patch("hamlet.backend.query.context.Context")
        @mock.patch("hamlet.backend.query.template")
        def wrapper(
            blueprint_mock, ContextClassMock, create_template_backend, *args, **kwargs
        ):
            with tempfile.TemporaryDirectory() as temp_cache_dir:

                ContextObjectMock = ContextClassMock()
                ContextObjectMock.md5_hash.return_value = str(
                    hashlib.md5(str(schemalist).encode()).hexdigest()
                )
                ContextObjectMock.cache_dir = temp_cache_dir

                blueprint_mock.run.side_effect = template_backend_run_mock(schemalist)

                return func(
                    blueprint_mock,
                    ContextClassMock,
                    create_template_backend,
                    *args,
                    **kwargs
                )

        return wrapper

    return decorator


schema_list = {
    "Stages": [
        {
            "Id": "StageId1",
            "Steps": [
                {
                    "Id": "StepId1",
                    "Parameters": {
                        "Schema": "Schema1",
                    },
                },
                {
                    "Id": "StepId2",
                    "Parameters": {
                        "Schema": "Schema2",
                    },
                },
            ],
        }
    ]
}


@mock_backend(schema_list)
def test_input_valid(blueprint_mock, ContextClassMock, create_template_backend):
    run_options_test(CliRunner(), create_schemas, ALL_VALID_OPTIONS, blueprint_mock.run)


@mock_backend(schema_list)
def test_input_validation(blueprint_mock, ContextClassMock, create_template_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        create_schemas,
        create_template_backend.run,
        {"-s": "Schema1", "-o": "."},
        [],
    )



import click
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options

from hamlet.backend.deploy import find_deployments, create_deployment, run_deployment


@click.command(
    "run-deployments", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-m",
    "--deployment-mode",
    default="update",
    help="The deployment mode to use for the deployment",
)
@click.option(
    "-l",
    "--deployment-group",
    default=".*",
    show_default=True,
    help="The deployment group pattern to match",
)
@click.option(
    "-u",
    "--deployment-unit",
    default=[".*"],
    show_default=True,
    multiple=True,
    help="The deployment unit pattern to match",
)
@click.option(
    "-s",
    "--deployment-state",
    type=click.Choice(
        [
            "deployed",
            "notdeployed",
            "orphaned",
        ],
        case_sensitive=False,
    ),
    default=["deployed", "notdeployed"],
    multiple=True,
    help="The states of deployments to include",
)
@click.option(
    "-d",
    "--district",
    default=["segment"],
    multiple=True,
    help="The districts to include deployments from",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
    ),
    help="the directory where the outputs will be saved",
)
@click.option(
    "--refresh-outputs/--no-refresh-outputs",
    default=True,
    help="Update outputs or use existing",
)
@click.option(
    "--confirm/--no-confirm",
    default=False,
    help="Confirm before executing each deployment",
)
@click.option(
    "--dryrun/--no-dryrun",
    default=False,
    help="Perform a dry run of the deployment before the run",
)
@exceptions.backend_handler()
@pass_options
def run_deployments(
    options,
    deployment_mode,
    deployment_group,
    deployment_unit,
    deployment_state,
    district,
    output_dir,
    refresh_outputs,
    confirm,
    dryrun,
    **kwargs,
):
    """
    Create and run deployments
    """
    deployments = find_deployments(
        deployment_mode,
        deployment_group,
        deployment_units=deployment_unit,
        deployment_states=deployment_state,
        districts=district,
        **options.opts,
    )

    if len(deployments) == 0:
        raise exceptions.CommandError("No deployments found that match pattern")

    for deployment in deployments:

        deployment_group = deployment["DeploymentGroup"]
        deployment_unit = deployment["DeploymentUnit"]
        deployment_state = deployment["CurrentState"]
        provider = deployment["DeploymentProvider"]

        click.echo("")
        click.secho(f"[*] {deployment_group}/{deployment_unit}", bold=True, fg="green")

        if deployment_state == "orphaned":
            click.secho(
                "[-] deployment has been orphaned, running orphan clean up",
                bold=False,
                fg="yellow",
            )

        click.echo("")

        if refresh_outputs:
            if deployment_state != "orphaned":
                create_deployment(
                    deployment_group,
                    deployment_unit,
                    deployment_mode,
                    output_dir,
                    **options.opts,
                )

        for operation in deployment["Operations"]:

            if dryrun:
                run_deployment(
                    provider,
                    deployment_group,
                    deployment_unit,
                    operation,
                    output_dir,
                    dryrun,
                    **options.opts,
                )

            if (
                confirm
                and click.confirm(
                    f"Start Deployment of {deployment_group}/{deployment_unit} ?"
                )
            ) or not confirm:
                run_deployment(
                    provider,
                    deployment_group,
                    deployment_unit,
                    operation,
                    output_dir,
                    **options.opts,
                )



import click
import os

from click_configfile import ConfigFileReader, SectionSchema, matches_section
from hamlet.env import HAMLET_GLOBAL_CONFIG
from hamlet.utils import ConfigParam


class ConfigSchema(object):
    """Schema for standard configuration."""

    @matches_section("profile:*")
    class Profile(SectionSchema):
        """Profile-specific configuration schema."""

        #: the root directory of the CMDB
        root_dir = ConfigParam(name="root_dir", type=click.Path())

        #: the name of the tenant
        tenant = ConfigParam(name="tenant", type=str)

        #: the name of the account
        account = ConfigParam(name="account", type=str)

        #: the name of the product
        product = ConfigParam(name="product", type=str)

        #: the name of the environment
        environment = ConfigParam(name="environment", type=str)

        #: the name of the segment
        segment = ConfigParam(name="segment", type=str)

        #: the name of the engine to use within this profile
        engine = ConfigParam(name="engine", type=str)


class ConfigReader(ConfigFileReader):
    """Reader for standard configuration."""

    config_files = ["config.ini", "config"]
    config_name = "standard"
    config_searchpath = []
    config_section_schemas = [ConfigSchema.Profile]

    @classmethod
    def select_config_schema_for(cls, section_name):
        section_schema = super(ConfigReader, cls).select_config_schema_for(section_name)
        for v in section_schema.__dict__:
            if isinstance(v, ConfigParam):
                v.ctx = cls
        return section_schema

    @classmethod
    def load_config(cls, opts, profile=None):
        """Load a configuration file into an options object."""

        if os.path.exists(HAMLET_GLOBAL_CONFIG.config_dir):
            if os.path.isdir(HAMLET_GLOBAL_CONFIG.config_dir):
                cls.config_searchpath.insert(0, HAMLET_GLOBAL_CONFIG.config_dir)

        config = cls.read_config()
        values = config.get("default", {})
        cls._load_values_into_opts(opts, values)

        if profile and profile != "default":
            values = config.get(f"profile:{profile}", {})
            cls._load_values_into_opts(opts, values)

        return values

    @staticmethod
    def _load_values_into_opts(opts, values):
        for k, v in values.items():
            if v is None:
                continue
            if isinstance(v, str):
                if v.startswith('"') or v.startswith("'"):
                    v = v[1:]
                if v.endswith('"') or v.endswith("'"):
                    v = v[:-1]
                if not v:
                    continue
            else:
                if v is None:
                    continue
            setattr(opts, k, v)


class Options:
    """Options object that holds config for the application."""

    def __init__(self, *args, **kwargs):
        """Initialise a new Options object."""
        super(Options, self).__init__(*args, **kwargs)
        self.opts = {}
        for k, v in kwargs:
            setattr(self, k, v)

    @staticmethod
    def get_config_reader():
        """Get the config reader class."""
        return ConfigReader

    def load_config_file(self, profile=None):
        """Load the config file."""
        config_cls = self.get_config_reader()
        return config_cls.load_config(self, profile=profile)

    @property
    def cli_config_dir(self):
        """The cli config dir"""
        return self._get_option("cli_config_dir")

    @cli_config_dir.setter
    def cli_config_dir(self, value):
        """Set the cli config dir"""
        self._set_option("cli_config_dir", value)

    @property
    def log_level(self):
        """Get the log_level setting"""
        return self._get_option("log_level")

    @log_level.setter
    def log_level(self, value):
        """Set the log_level setting"""
        self._set_option("log_level", value)

    @property
    def root_dir(self):
        """Get the root dir setting"""
        return self._get_option("root_dir")

    @root_dir.setter
    def root_dir(self, value):
        """Set the root_dir setting"""
        self._set_option("root_dir", value)

    @property
    def tenant(self):
        """Get the tenant setting"""
        return self._get_option("tenant")

    @tenant.setter
    def tenant(self, value):
        """Set the tenant setting"""
        self._set_option("tenant", value)

    @property
    def account(self):
        """Get the account setting"""
        return self._get_option("account")

    @account.setter
    def account(self, value):
        """Set the account setting"""
        self._set_option("account", value)

    @property
    def product(self):
        """Get the product setting"""
        return self._get_option("product")

    @product.setter
    def product(self, value):
        """Set the product setting"""
        self._set_option("product", value)

    @property
    def environment(self):
        """Get the environment setting"""
        return self._get_option("environment")

    @environment.setter
    def environment(self, value):
        """Set the environment setting"""
        self._set_option("environment", value)

    @property
    def segment(self):
        """Get the segment setting"""
        return self._get_option("segment")

    @segment.setter
    def segment(self, value):
        """Set the segment setting"""
        self._set_option("segment", value)

    @property
    def engine(self):
        """Get the engine to use for the district"""
        return self._get_option("engine")

    @engine.setter
    def engine(self, value):
        """Set the engine setting"""
        self._set_option("engine", value)

    @property
    def generation_framework(self):
        """Get the generation_framework setting"""
        return self._get_option("generation_framework")

    @generation_framework.setter
    def generation_framework(self, value):
        """Set the generation_framework setting"""
        self._set_option("generation_framework", value)

    @property
    def generation_provider(self):
        """Get the generation_provider setting"""
        return self._get_option("generation_provider")

    @generation_provider.setter
    def generation_provider(self, value):
        """Set the generation_provider setting"""
        self._set_option("generation_provider", value)

    @property
    def generation_input_source(self):
        """Get the generation_input_source setting"""
        return self._get_option("generation_input_source")

    @generation_input_source.setter
    def generation_input_source(self, value):
        """Set the generation_input_source setting"""
        self._set_option("generation_input_source", value)

    def _get_option(self, name, default=None):
        """Get value for an option."""
        value = self.opts.get(name)
        if value is None:
            return default
        return value

    def _set_option(self, name, value, allow_clear=False):
        """Set value for an option."""
        if not allow_clear:
            # Prevent clears if value was set
            try:
                current_value = self._get_option(name)
                if value is None and current_value is not None:
                    return
            except AttributeError:
                pass
        self.opts[name] = value


pass_options = click.make_pass_decorator(Options, ensure=True)



# NOTE: This file must remain valid python file in order to perform tests on it.
# NOTE: Imports can't be used inside a template block because all code will be merged into a single file.
# NOTE: The class is wrapped into function in order to make it testable. This adds an ability to provide parent classes
# at runtime, otherwise, the module will raise NameError


def JSONStructure(JSONValidator):
    class JSONStructure(JSONValidator):
        def __match(self, path, target):
            def validator():
                value = self._get_value_by_json_path(path)
                assert value == target, "{} doesn't match {}".format(path, target)

            validator.__name__ = "{} == {}".format(path, target)
            return validator

        def __len(self, path, target):
            def validator():
                value = self._get_value_by_json_path(path)
                assert len(value) == target, "{}.length={}!={}".format(
                    path, len(value), target
                )

            validator.__name__ = "{}.length == {}".format(path, target)
            return validator

        def __exists(self, path):
            def validator():
                return self._get_value_by_json_path(path)

            validator.__name__ = "{} exists?".format(path)
            return validator

        def __not_empty(self, path):
            def validator():
                value = self._get_value_by_json_path(path)
                assert_text = "{} is empty".format(path)
                assert value is not None, assert_text
                if isinstance(value, (str, dict, list)):
                    assert len(value) > 0, assert_text

            validator.__name__ = "not empty {}?".format(path)
            return validator

        def match(self, path, target):
            self._validators.append(self.__match(path, target))
            return self

        def len(self, path, target):
            self._validators.append(self.__len(path, target))
            return self

        def exists(self, path):
            self._validators.append(self.__exists(path))
            return self

        def not_empty(self, path):
            self._validators.append(self.__not_empty(path))
            return self

    return JSONStructure



import os
import json
import pathlib


class Search:
    @staticmethod
    def split_path(path):
        parts = os.path.normpath(path).split(os.path.sep)
        if len(parts) > 0 and parts[0] == "":
            parts[0] = os.path.sep
        return parts

    @staticmethod
    def exists(directory, name, up=0):
        if up < 0:
            raise ValueError("up parameter must be >= 0")
        search_dir_parts = Search.split_path(directory)
        up = min(len(search_dir_parts), up)
        up = len(search_dir_parts) - up
        search_path = os.path.join(*search_dir_parts[:up], name)
        return search_path if os.path.exists(search_path) else None

    @staticmethod
    def isfile(directory, name, up=0):
        filename = Search.exists(directory, name, up)
        return filename if filename is not None and os.path.isfile(filename) else None

    @staticmethod
    def isdir(directory, name, up=0):
        dirname = Search.exists(directory, name, up)
        return dirname if dirname is not None and os.path.isdir(dirname) else None

    @staticmethod
    def upwards(directory, name):
        parts = Search.split_path(directory)
        for up in range(len(parts)):
            up = len(parts) - up
            search_path = os.path.join(*parts[:up], name)
            if os.path.exists(search_path):
                return search_path
        return None

    @staticmethod
    def downwards(directory, name):
        found = []
        for root, dirs, files in os.walk(directory, topdown=True):
            for filename in files:
                if filename == name:
                    found.append(os.path.join(root, filename))
            for dirname in dirs:
                if dirname == name:
                    found.append(os.path.join(root, dirname))
        return found

    @staticmethod
    def basename(path, up=0):
        parts = Search.split_path(path)
        up = min(len(parts), up)
        up = len(parts) - up - 1
        return parts[up]

    @staticmethod
    def parent(path, up=0):
        parts = Search.split_path(path)
        up = min(len(parts), up)
        up = len(parts) - up
        return os.path.join(*parts[:up])

    @staticmethod
    def cut(path, prefix="", suffix=""):
        path = Search.split_path(path)
        if prefix:
            prefix = Search.split_path(prefix)
            if path[: len(prefix)] == prefix:
                path = path[len(prefix) :]
        if suffix:
            suffix = Search.split_path(suffix)
            if path[-len(suffix) :] == suffix:
                path = path[: -len(suffix)]
        return os.path.join(*path)

    @staticmethod
    def match(*patterns, root=None):
        root = root or os.getcwd()
        root = pathlib.Path(root)
        matches = []
        for pattern in patterns:
            matches_gen = root.glob(pattern)
            matches += [str(match) for match in matches_gen]
        # removing duplicates
        return list(set(matches))

    def match_dirs(*patterns, root=None, include_file_dirs=False):
        matches = Search.match(*patterns, root=root)
        filtered_matches = []
        for match in matches:
            if os.path.isfile(match):
                if include_file_dirs:
                    filtered_matches.append(os.path.dirname(match))
            else:
                filtered_matches.append(match)
        # removing duplicates
        return list(set(filtered_matches))

    def match_files(*patterns, root=None):
        matches = Search.match(*patterns, root=root)
        return [match for match in matches if os.path.isfile(match)]

    def list_all(dirname):
        return [name for name in os.listdir(dirname)]

    def list_files(dirname):
        return [
            name
            for name in os.listdir(dirname)
            if os.path.isfile(os.path.join(dirname, name))
        ]

    def list_dirs(dirname):
        return [
            name
            for name in os.listdir(dirname)
            if os.path.isdir(os.path.join(dirname, name))
        ]


class ContextSearch:
    def __init__(self, cwd):
        self.cwd = cwd

    @property
    def cwd(self):
        return self.__cwd

    @cwd.setter
    def cwd(self, value):
        if not os.path.isabs(value):
            raise ValueError("{} is not abs path".format(value))
        if not os.path.exists(value):
            raise ValueError("{} does not exist".format(value))
        if not os.path.isdir(value):
            raise ValueError("{} is not directory".format(value))
        self.__cwd = value

    def exists(self, name, up=0):
        return Search.exists(self.cwd, name, up=up)

    def isfile(self, name, up=0):
        return Search.isfile(self.cwd, name, up=up)

    def isdir(self, name, up=0):
        return Search.isdir(self.cwd, name, up=up)

    def upwards(self, name):
        return Search.upwards(self.cwd, name)

    def downwards(self, name):
        return Search.downwards(self.cwd, name)

    def basename(self, up=0):
        return Search.basename(self.cwd, up=up)

    def parent(self, up=0):
        return Search.parent(self.cwd, up=up)


# TODO: Needs to be splitted into several subclasses like JSONFile, TXTFile, ByteFile
# Then File may be used as an automated type resolver.
class File:
    def __init__(self, path):
        self.path = path
        self.data = None
        name, ext = os.path.splitext(self.path)
        self.ext = ext

    def load(self):
        if self.data is not None:
            return self.data
        if self.ext == ".json":
            with open(self.path, "rb") as f:
                self.data = json.load(f)
        else:
            with open(self.path, "rt") as f:
                self.data = f.read()
        return self.data

    def reload(self):
        self.data = None
        return self.load()

    def write(self):
        if self.ext == ".json":

            def serialize():
                return json.dumps(self.data)

        else:

            def serialize():
                return self.data

        with open(self.path, "wt+") as f:
            f.write(serialize())

    def __getitem__(self, key):
        try:
            self.load()
            return self.data[key]
        except TypeError as e:
            raise TypeError("Unstructured data") from e


class Directory:
    def __init__(self, path):
        self.path = path

    def __getitem__(self, key):
        path = os.path.join(self.path, key)
        if not os.path.exists(path):
            raise KeyError("Path {} does not exist".format(path))
        if os.path.isfile(path):
            return File(path)
        else:
            return self.__class__(os.path.join(self.path, key))

    def __iter__(self):
        for path in os.listdir(self.path):
            fullpath = os.path.join(self.path, path)
            if os.path.isfile(fullpath):
                yield File(fullpath)
            elif os.path.isdir(fullpath):
                yield self.__class__(fullpath)
            else:
                yield fullpath



from hamlet.backend.common import runner


def run(
    credential_email=None,
    crypto_file=None,
    credential_id=None,
    credential_path=None,
    credential_secret=None,
    visible=None,
    credential_type=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    **kwargs,
):
    options = {
        "-e": credential_email,
        "-f": crypto_file,
        "-i": credential_id,
        "-n": credential_path,
        "-v": visible,
        "-y": credential_type,
        "-s": credential_secret,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("manageCredentialCrypto.sh", [], options, env, _is_cli)



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.manage.deployment import deployment as manage_deployment
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "unit"
ALL_VALID_OPTIONS["!-l,--deployment-group"] = "group"
ALL_VALID_OPTIONS["-d,--delete"] = [True, False]
ALL_VALID_OPTIONS["-g,--resource-group"] = "group"
ALL_VALID_OPTIONS["-i,--deployment-initiate"] = [True, False]
ALL_VALID_OPTIONS["-m,--deployment-monitor"] = [True, False]
ALL_VALID_OPTIONS["-r,--region"] = "region"
ALL_VALID_OPTIONS["-s,--deployment-scope"] = ["subscription", "resourceGroup"]
ALL_VALID_OPTIONS["-w,--deployment-wait"] = 10
ALL_VALID_OPTIONS["-z,--deployment-unit-subset"] = "subset"


@mock.patch("hamlet.command.manage.deployment.manage_deployment_backend")
def test_input_valid(manage_deployment_backend):
    run_options_test(
        CliRunner(), manage_deployment, ALL_VALID_OPTIONS, manage_deployment_backend.run
    )


@mock.patch("hamlet.command.manage.deployment.manage_deployment_backend")
def test_input_validation(manage_deployment_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        manage_deployment,
        manage_deployment_backend.run,
        {"-u": "unit", "-l": "group"},
        [("-s", "badscopevalue", "resourceGroup"), ("-w", "not_an_int", 10)],
    )



import collections
import tempfile
import hashlib
import json
import os

from unittest import mock
from click.testing import CliRunner
from hamlet.command.visual import draw_diagrams as draw_diagrams
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()

ALL_VALID_OPTIONS["!-i,--diagram-id"] = ["Diagram1"]
ALL_VALID_OPTIONS["!-d,--asset-dir"] = "./"
ALL_VALID_OPTIONS["-s,--src-dir"] = "./"


def template_backend_run_mock(data):
    def run(
        entrance="diagraminfo",
        output_filename="diagraminfo.json",
        deployment_mode=None,
        output_dir=None,
        generation_input_source=None,
        generation_provider=None,
        generation_framework=None,
        log_level=None,
        root_dir=None,
        tenant=None,
        account=None,
        product=None,
        environment=None,
        segment=None,
    ):
        os.makedirs(output_dir, exist_ok=True)
        unitlist_filename = os.path.join(output_dir, output_filename)
        with open(unitlist_filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(unitlist=None):
    def decorator(func):
        @mock.patch("hamlet.command.visual.create_diagram_backend")
        @mock.patch("hamlet.command.visual.create_template_backend")
        @mock.patch("hamlet.backend.query.context.Context")
        @mock.patch("hamlet.backend.query.template")
        def wrapper(
            blueprint_mock,
            ContextClassMock,
            create_template_backend,
            create_diagram_backend,
            *args,
            **kwargs
        ):
            with tempfile.TemporaryDirectory() as temp_cache_dir:

                ContextObjectMock = ContextClassMock()
                ContextObjectMock.md5_hash.return_value = str(
                    hashlib.md5(str(unitlist).encode()).hexdigest()
                )
                ContextObjectMock.cache_dir = temp_cache_dir

                blueprint_mock.run.side_effect = template_backend_run_mock(unitlist)

                return func(blueprint_mock, ContextClassMock, *args, **kwargs)

        return wrapper

    return decorator


diagram_list = {
    "Diagrams": [
        {
            "Id": "Diagram1",
        }
    ]
}


@mock_backend(diagram_list)
def test_input_valid(create_template_backend, create_diagram_backend):
    run_options_test(
        CliRunner(), draw_diagrams, ALL_VALID_OPTIONS, create_template_backend.run
    )


@mock_backend(diagram_list)
def test_input_validation(create_template_backend, create_diagram_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        draw_diagrams,
        create_template_backend.run,
        {"-i": "overview", "-d": "./"},
        [],
    )



import hashlib
import os

import pathlib
import json

from abc import ABC, abstractmethod
from hamlet.backend.container_registry import ContainerRepository


class EngineSourceInterface(ABC):
    """
    The engine source implements the retrevel of hamlet component artefacts
    Source classes are based on the process require to retrieve the arefacts
    """

    def __init__(self, name, description=""):
        self.name = name
        self.description = description
        self.env_path = None

    @abstractmethod
    def pull(self, dst_dir):
        """
        Pull the source to a local directory
        Must return a status object describing the source state
        """
        return EngineSourcePullState(
            name=self.name, type=self.__class__.__name__, digest=self.name
        )

    @property
    @abstractmethod
    def digest(self):
        """
        Return a digest of the source to use for verification and version updates
        """
        raise NotImplementedError


class ShimPathEngineSource(EngineSourceInterface):
    """
    Sets a standard source path to use for shim based providers
    """

    def pull(self, dst_dir):
        if not os.path.isdir(dst_dir):
            os.makedirs(dst_dir)

        return EngineSourcePullState(
            name=self.name, type=self.__class__.__name__, digest=self.name
        )

    @property
    def digest(self):
        return self.name


class LocalDirectoryEngineSource(EngineSourceInterface):
    """
    An engine source that is available from the local file system
    """

    def __init__(self, name, description, env_path):
        super().__init__(name, description=description)
        self.env_path = env_path

    def pull(self, dst_dir):
        if not os.path.isdir(dst_dir):
            os.makedirs(dst_dir)

        return EngineSourcePullState(
            name=self.name,
            type=self.__class__.__name__,
            digest=hashlib.sha256(bytes(self.env_path, encoding="utf8")).hexdigest(),
            source_metadata={"local_path": self.env_path},
        )

    @property
    def digest(self):
        return hashlib.sha256(bytes(self.env_path, encoding="utf8")).hexdigest()


class ContainerEngineSource(EngineSourceInterface):
    """
    Container based engine source which uses docker registries to pull content
    """

    def __init__(
        self,
        name,
        description,
        registry_url,
        repository,
        tag,
        username=None,
        password=None,
    ):
        super().__init__(name, description)

        self.registry_url = registry_url
        self.repository = repository
        self.tag = tag

        self.username = username
        self.password = password

        self._container_repository = ContainerRepository(
            registry_url=self.registry_url,
            repository=self.repository,
            username=self.username,
            password=self.password,
        )

    def pull(self, dst_dir):

        pull_digest = self._container_repository.pull(self.tag, dst_dir)

        return EngineSourcePullState(
            name=self.name,
            type=self.__class__.__name__,
            digest=pull_digest,
            source_metadata={
                "registry_url": self.registry_url,
                "repository": self.repository,
                "tag": self.tag,
            },
            build_metadata=self._get_build_details(dst_dir),
        )

    @property
    def digest(self):
        return self._container_repository.get_tag_digest(self.tag)

    def _get_build_details(self, dst_dir):

        engine_source = ".hamlet/engine_source.json"
        if dst_dir is not None:
            engine_source_paths = pathlib.Path(dst_dir).glob(f"**/{engine_source}")
            build_sources = {}
            for engine_source_path in engine_source_paths:
                with open(engine_source_path, "r") as file:
                    source_name = str(engine_source_path)[
                        len(dst_dir) : len(str(engine_source_path)) - len(engine_source)
                    ]
                    build_sources[source_name] = json.load(file)

            return build_sources


class EngineSourcePullState(dict):
    """
    Provides details about the source that was pulled
    """

    def __init__(self, name, type, digest, source_metadata=None, build_metadata=None):
        self.name = name
        self.type = type
        self.digest = digest
        self.source_metadata = source_metadata
        self.build_metadata = build_metadata



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.entrance import invoke_entrance as invoke_entrance
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()

ALL_VALID_OPTIONS["!-e,--entrance"] = "entrance"
ALL_VALID_OPTIONS["-l,--deployment-group"] = "deployment_group"
ALL_VALID_OPTIONS["-u,--deployment-unit"] = "deployment_unit"
ALL_VALID_OPTIONS["-z,--deployment-unit-subset"] = "deployment_unit_subset"
ALL_VALID_OPTIONS["-d,--deployment-mode"] = "deployment_mode"
ALL_VALID_OPTIONS["-x,--disable-output-cleanup"] = [True]
ALL_VALID_OPTIONS["-o,--output-dir"] = "output_dir"
ALL_VALID_OPTIONS["-c,--config-ref"] = "config_ref"
ALL_VALID_OPTIONS["-q,--request-ref"] = "request_ref"


@mock.patch("hamlet.command.entrance.create_template_backend")
def test_input_valid(create_template_backend):
    run_options_test(
        CliRunner(), invoke_entrance, ALL_VALID_OPTIONS, create_template_backend.run
    )


@mock.patch("hamlet.command.entrance.create_template_backend")
def test_input_validation(create_template_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner, invoke_entrance, create_template_backend.run, {"-e": "blueprint"}, []
    )



from hamlet.command import root as cli

#  to avoid mock.patch path interpretation issuses
from .generate import geneate as generate_cmd
from .run import run as run_cmd


@cli.group("test", context_settings=dict(max_content_width=240))
def group():
    """
    Tests stuff
    """


group.add_command(generate_cmd)
group.add_command(run_cmd)



from hamlet.backend.common import runner


def run(
    container_id=None,
    delay=None,
    env_name=None,
    component=None,
    component_instance=None,
    component_version=None,
    tier=None,
    value=None,
    task=None,
    instance=None,
    version=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    **kwargs
):
    options = {
        "-c": container_id,
        "-d": delay,
        "-e": env_name,
        "-i": component,
        "-j": component_instance,
        "-k": component_version,
        "-t": tier,
        "-v": value,
        "-w": task,
        "-x": instance,
        "-y": version,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("runTask.sh", [], options, env, _is_cli)



import json


class Environment:
    def __init__(self, data=None):
        self._data = dict()
        if data:
            self._data.update(data)

    def __getattribute__(self, name):
        if name.startswith("_"):
            return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(self, "_data").get(name, "")

    def __setattr__(self, name, value):
        if name.startswith("_"):
            object.__setattr__(self, name, value)
        else:
            object.__getattribute__(self, "_data")[name] = value

    def __getitem__(self, key):
        return object.__getattribute__(self, "_data").get(key, "")

    def __setitem__(self, key, value):
        object.__getattribute__(self, "_data")[key] = value

    def __str__(self):
        return str(json.dumps(self._data, indent=4, sort_keys=True))



from hamlet.backend.common import runner


def run(
    release_mode=None,
    deployment_mode=None,
    log_level=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    context_credentials=None,
    _is_cli=False,
    env={},
):
    env = {
        "AUTOMATION_LOG_LEVEL": log_level,
        "AUTOMATION_CONTEXT_CREDENTIALS": context_credentials,
        **env,
    }
    opts = {
        "-t": tenant,
        "-a": account,
        "-p": product,
        "-e": environment,
        "-s": segment,
        "-d": deployment_mode,
        "-r": release_mode,
    }
    runner.run(
        "setContext.sh",
        [],
        opts,
        env,
        _is_cli,
        script_base_path_env="AUTOMATION_BASE_DIR",
    )



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.lambda_func import lambda_func as run_lambda_func
from tests.unit.command.test_option_generation import run_options_test


ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "unit"
ALL_VALID_OPTIONS["!-f,--function-id"] = "function_id"
ALL_VALID_OPTIONS["-i,--input-payload"] = "payload"
ALL_VALID_OPTIONS["-l,--include-log-tail"] = [True, False]


@mock.patch("hamlet.command.run.lambda_func.run_lambda_func_backend")
def test_input_valid(run_lambda_func_backend):
    run_options_test(
        CliRunner(), run_lambda_func, ALL_VALID_OPTIONS, run_lambda_func_backend.run
    )



import tempfile
import os
import hashlib
from unittest import mock

from hamlet.backend.engine import EngineStore
from hamlet.backend.engine.common import ENGINE_GLOBAL_NAME
from hamlet.backend.engine.engine import Engine
from hamlet.backend.engine.loaders.base import GlobalEngineLoader, InstalledEngineLoader
from hamlet.backend.engine.loaders.user import UserDefinedEngineLoader
from hamlet.backend.engine.loaders.unicycle import UnicycleEngineLoader
from hamlet.backend.engine.engine_part import CoreEnginePart
from hamlet.backend.engine.engine_source import ShimPathEngineSource


def mock_container_registry():
    def decorator(func):
        @mock.patch("hamlet.backend.engine.engine_source.ContainerRepository")
        def wrapper(container_repository, *args, **kwargs):

            container_repository.return_value.get_tag_digest.return_value = "digest[1]"
            container_repository.return_value.get_tag_manifest.return_value = {
                "schemaVersion": 2,
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "config": {"digest": "config_digest[1]"},
                "layers": [],
            }
            container_repository.return_value.pull.return_value = "digest[1]"

            return func(container_repository, *args, **kwargs)

        return wrapper

    return decorator


def test_global_engine_loading():
    """
    Tests that the basic global engine is loaded and can be found
    """
    with tempfile.TemporaryDirectory() as store_dir:
        engine_store = EngineStore(store_dir=store_dir)

        engine_store.local_engine_loaders = [GlobalEngineLoader()]

        assert len(engine_store.get_engines()) == 1

        global_engine = engine_store.get_engine(ENGINE_GLOBAL_NAME)

        assert global_engine.name == ENGINE_GLOBAL_NAME

        global_engine.install()

        assert os.path.isfile(global_engine.engine_state_file)

        generation_dir_path = os.path.join(
            global_engine.install_path, "shim/executor-bash/cli"
        )
        assert global_engine.environment["GENERATION_DIR"] == generation_dir_path

        automation_dir_path = os.path.join(
            global_engine.install_path, "shim/executor-bash/automation/jenkins/aws"
        )
        assert global_engine.environment["AUTOMATION_DIR"] == automation_dir_path


def test_installed_engine_loading():
    """
    Tests that we can discover engines that have been installed
    and that they can be discovered by the InstalledEngine basic loader
    """
    with tempfile.TemporaryDirectory() as store_dir:
        engine_store = EngineStore(store_dir=store_dir)

        engine_store.local_engine_loaders = [
            InstalledEngineLoader(engine_store.engine_dir)
        ]

        assert len(engine_store.get_engines()) == 0

        """
        Engine is created in a directory which matches the engine_store engine search location
        The engine isn't loaded through an engine loader yet
        """
        manually_installed_engine = Engine(
            name="installed_engine", description="an installed engine"
        )
        manually_installed_engine.engine_dir = engine_store.engine_dir

        manually_installed_engine.sources = [
            ShimPathEngineSource(name="shim_source", description="A basic shim engine")
        ]

        manually_installed_engine.parts = [
            CoreEnginePart(source_name="shim_source", source_path="")
        ]

        manually_installed_engine.install()
        assert os.path.isfile(manually_installed_engine.engine_state_file)

        """
        Use the Installed loader to discover the manually installed engine
        """
        discovered_engine = engine_store.get_engine("installed_engine")
        assert discovered_engine.name == "installed_engine"

        generation_engine_path = os.path.join(
            discovered_engine.install_path, "shim_source"
        )
        assert (
            discovered_engine.environment["GENERATION_ENGINE_DIR"]
            == generation_engine_path
        )


@mock_container_registry()
def test_unicycle_engine_loading(container_repository):
    """
    tests that we can run source loading from a container
    """

    with tempfile.TemporaryDirectory() as store_dir:
        engine_store = EngineStore(store_dir=store_dir)

        engine_store.external_engine_loaders = [UnicycleEngineLoader()]

        unicycle_engine = engine_store.find_engine("unicycle")
        unicycle_engine.install()

        assert unicycle_engine.name == "unicycle"

        container_digests = ["digest[1]"] * len(unicycle_engine.sources)
        expected_digest = (
            "sha256:"
            + hashlib.sha256(":".join(container_digests).encode("utf-8")).hexdigest()
        )
        assert unicycle_engine.digest == expected_digest


@mock.patch("hamlet.backend.engine.loaders.user.HAMLET_GLOBAL_CONFIG")
def test_user_engine_loading(global_config):
    """
    Tests the user engine loading process
    """

    with tempfile.TemporaryDirectory() as test_dir:
        with tempfile.TemporaryDirectory() as store_dir:
            with tempfile.TemporaryDirectory() as cli_config_dir:

                type(global_config).config_dir = mock.PropertyMock(
                    return_value=cli_config_dir
                )

                with open(os.path.join(cli_config_dir, "engine"), "w") as engine_file:
                    engine_file.write(
                        (
                            "[engine:test]\n"
                            "\tdescription = test engine\n"
                            "\tsources = test_dir\n"
                            "\tparts = bash\n"
                            "\n"
                            "[engine_source:test_dir]\n"
                            "\ttype = local_dir\n"
                            f"\tlocal_dir_path = {test_dir}\n"
                            "\n"
                            "[engine_part:bash]\n"
                            "\ttype = executor-bash\n"
                            "\tsource_name = test_dir\n"
                        )
                    )

                engine_store = EngineStore(store_dir=store_dir)

                engine_store.external_engine_loaders = [UserDefinedEngineLoader()]

                local_test_engine = engine_store.find_engine("test")
                local_test_engine.install()

                assert local_test_engine.name == "test"



import os
import json
import tempfile
import pytest
from hamlet.backend.common.context import (
    Context,
    NoLevelFileError,
    NoRootFileError,
    NoTenantFoundError,
    NoAccountsFoundError,
    MultipleTenantsFoundError,
    MultipleAccountsFoundError,
    SpecifiedAccountNotFoundError,
    AccountLevel,
    TenantLevel,
    ProductLevel,
    EnvironmentLevel,
    SegmentLevel,
)


def test_context_basics():
    with tempfile.TemporaryDirectory() as temp_dir:

        def cmdb(*paths):
            return os.path.join(temp_dir, "cmdb", *paths)

        # setting fake cmdb
        os.mkdir(cmdb())
        open(cmdb("root.json"), "w").close()
        os.makedirs(cmdb("accounts"))
        os.makedirs(cmdb("accounts", "tenant"))
        os.makedirs(cmdb("accounts", "tenant-second"))
        os.makedirs(cmdb("accounts", "account-name-1", "config"))
        os.makedirs(cmdb("accounts", "account-name-2", "config"))
        open(cmdb("accounts", "tenant", "tenant.json"), "w").close()
        open(cmdb("accounts", "tenant-second", "tenant.json"), "w").close()
        os.makedirs(cmdb("product", "config"))
        os.makedirs(cmdb("product", "config", "solutions", "environment"))
        os.makedirs(cmdb("product", "config", "solutions", "environment", "segment"))
        with open(
            cmdb("product", "config", "solutions", "environment", "environment.json"),
            "wt+",
        ) as f:
            json.dump({"Name": "environment_name"}, f)
        with open(
            cmdb(
                "product",
                "config",
                "solutions",
                "environment",
                "segment",
                "segment.json",
            ),
            "wt+",
        ) as f:
            json.dump({"Name": "segment_name"}, f)
        with open(cmdb("product", "config", "product.json"), "wt+") as f:
            json.dump({"Name": "product_name"}, f)
        with open(
            cmdb("accounts", "account-name-1", "config", "account.json"), "wt+"
        ) as f:
            json.dump({"Name": "account-name-1"}, f)
        with open(
            cmdb("accounts", "account-name-2", "config", "account.json"), "wt+"
        ) as f:
            json.dump({"Name": "account-name-2"}, f)

        context = Context(cmdb())
        assert context.root_dir == cmdb()

        with pytest.raises(NoRootFileError):
            Context(temp_dir)

        with pytest.raises(MultipleTenantsFoundError):
            AccountLevel(cmdb("accounts", "account-name-1"))

        os.remove(cmdb("accounts", "tenant-second", "tenant.json"))
        account_level = AccountLevel(cmdb("accounts", "account-name-1"))
        assert account_level.props["Name"] == "account-name-1"
        assert account_level.tenant_dir == cmdb("accounts", "tenant")
        assert account_level.level_file_path == cmdb(
            "accounts", "account-name-1", "config", "account.json"
        )

        tenant_level = TenantLevel(cmdb("accounts", "tenant"))
        assert tenant_level.level_file_path == cmdb("accounts", "tenant", "tenant.json")

        with pytest.raises(NoLevelFileError):
            ProductLevel(
                cmdb("accounts", "account-name-1"),
                config=dict(account="account-name-1"),
            )

        with pytest.raises(MultipleAccountsFoundError):
            ProductLevel(cmdb("product"))

        product_level = ProductLevel(
            cmdb("product"), config=dict(account="account-name-1")
        )
        assert product_level.tenant_dir == cmdb("accounts", "tenant")
        assert product_level.account_dir == cmdb("accounts", "account-name-1")

        product_level = ProductLevel(
            cmdb("product"), config=dict(account="account-name-2")
        )
        assert product_level.tenant_dir == cmdb("accounts", "tenant")
        assert product_level.account_dir == cmdb("accounts", "account-name-2")

        with pytest.raises(SpecifiedAccountNotFoundError):
            ProductLevel(cmdb("product"), config=dict(account="account-name-3"))

        environment_level = EnvironmentLevel(
            cmdb("product", "config", "solutions", "environment"),
            config=dict(account="account-name-1"),
        )
        assert environment_level.tenant_dir == cmdb("accounts", "tenant")
        assert environment_level.account_dir == cmdb("accounts", "account-name-1")
        assert environment_level.props["Environment"] == "environment"

        segment_level = SegmentLevel(
            cmdb("product", "config", "solutions", "environment", "segment"),
            config=dict(account="account-name-2"),
        )
        assert segment_level.tenant_dir == cmdb("accounts", "tenant")
        assert segment_level.account_dir == cmdb("accounts", "account-name-2")
        assert segment_level.props["Environment"] == "environment"
        assert segment_level.props["Segment"] == "segment"

        os.remove(cmdb("accounts", "account-name-1", "config", "account.json"))

        product_level = ProductLevel(
            cmdb("product"), config=dict(account="account-name-2")
        )
        assert product_level.tenant_dir == cmdb("accounts", "tenant")
        assert product_level.account_dir == cmdb("accounts", "account-name-2")

        with pytest.raises(SpecifiedAccountNotFoundError):
            ProductLevel(cmdb("product"), config=dict(account="account-name-1"))

        product_level = ProductLevel(cmdb("product"))
        assert product_level.tenant_dir == cmdb("accounts", "tenant")
        assert product_level.account_dir == cmdb("accounts", "account-name-2")

        os.remove(cmdb("accounts", "account-name-2", "config", "account.json"))

        with pytest.raises(NoAccountsFoundError):
            ProductLevel(cmdb("product"))

        os.remove(cmdb("accounts", "tenant", "tenant.json"))
        with pytest.raises(NoTenantFoundError):
            ProductLevel(cmdb("product"))



from hamlet.backend.common import runner


def run(
    release_mode=None,
    acceptance_tag=None,
    deployment_unit=None,
    code_commit=None,
    code_tag=None,
    image_format=None,
    registry_scope=None,
    log_level=None,
    _is_cli=False,
    env={},
):
    env = {
        "AUTOMATION_LOG_LEVEL": log_level,
        "DEPLOYMENT_UNIT_LIST": deployment_unit,
        "CODE_COMMIT": code_commit,
        "CODE_TAG": code_tag,
        "IMAGE_FORMAT": image_format,
        "REGISTRY_SCOPE": registry_scope,
        "RELEASE_MODE": release_mode,
        "ACCEPTANCE_TAG": acceptance_tag,
        **env,
    }
    runner.run(
        "confirmBuilds.sh", [], {}, env, _is_cli, script_base_path_env="AUTOMATION_DIR"
    )



from . import step_1_install_engine
from . import step_2_create_cmdb
from . import step_3_create_account_templates
from . import step_4_create_segment_templates


def test_tenant_and_product_setup(engine, clear_engine, cmdb, clear_cmdb):
    step_1_install_engine.run(engine, clear_engine)
    step_2_create_cmdb.run(cmdb, clear_cmdb)
    step_3_create_account_templates.run(cmdb, engine)
    step_4_create_segment_templates.run(cmdb, engine)



import os
import click
import json

from tabulate import tabulate

from hamlet.command import root as cli
from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.backend import query as query_backend


def query_info_output(options, query, query_params=None, sub_query_text=None):
    query_args = {
        **options.opts,
        "generation_entrance": "info",
        "output_filename": "info.json",
        "use_cache": False,
    }
    query_result = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        query_text=query,
        query_params=query_params,
        sub_query_text=sub_query_text
    )

    return query_result


@cli.group("layer", context_settings=dict(max_content_width=240))
def group():
    """
    Layers provided in the CMDB
    """
    pass


def layer_types_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Type"]),
                wrap_text(row["ReferenceLookupType"]),
                wrap_text(row["Description"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Type", "ReferenceLookupType", "Description"],
        tablefmt="github",
    )


@group.command(
    "list-layer-types", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="A query to filter out the results")
@json_or_table_option(layer_types_table)
@exceptions.backend_handler()
@pass_options
def list_layer_types(options, query):
    """
    Lists the types of layers available
    """

    LIST_LAYER_TYPES_QUERY = (
        "LayerTypes[]"
        ".{"
        "Type:Type,"
        "ReferenceLookupType:ReferenceLookupType,"
        "Description:Description"
        "}"
    )

    return query_info_output(options, LIST_LAYER_TYPES_QUERY, None, query)


@group.command(
    "describe-layer-type",
    short_help="",
    context_settings=dict(max_content_width=240),
)
@click.option("-t", "--type", required=True, help="The type of the layer to describe")
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def describe_layer_type(options, type, query):
    """
    Describes a specific layer type
    """
    DESCRIBE_LAYER_TYPE_QUERY = "LayerTypes[?Type=={type}] | [0]"

    click.echo(
        json.dumps(
            query_info_output(
                options, DESCRIBE_LAYER_TYPE_QUERY, {"type": type}, query
            ),
            indent=2,
        )
    )


@group.command(
    "query-layer-types", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def query_layer_types(options, query):
    """
    Query across all layer types
    """
    click.echo(
        json.dumps(query_info_output(options, "LayerTypes[]", None, query), indent=2)
    )


def layer_list_data_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Name"]),
                wrap_text(row["Id"]),
                wrap_text(row["Type"]),
                wrap_text(row["Active"]),
            ]
        )
    return tabulate(
        tablerows, headers=["Name", "Id", "Type", "Active"], tablefmt="github"
    )


@group.command(
    "list-layers", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="A query to filter out the results")
@json_or_table_option(layer_list_data_table)
@exceptions.backend_handler()
@pass_options
def list_layers(options, query):
    """
    Lists the layers available
    """

    LIST_LAYERS_QUERY = "LayerData[].{Name:Name,Id:Id,Type:Type,Active:Active}"

    return query_info_output(options, LIST_LAYERS_QUERY, {}, query)


@group.command(
    "describe-layer", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-t", "--type", required=True, help="The type of the layer to describe")
@click.option("-n", "--name", required=True, help="The name of the layer to describe")
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def describe_layer(options, type, name, query):
    """
    Describes a specific layer
    """
    DESCRIBE_LAYER_QUERY = "LayerData[?Type=={type} && Name=={name}] | [0]"

    click.echo(
        json.dumps(
            query_info_output(
                options, DESCRIBE_LAYER_QUERY, {"type": type, "name": name}, query
            ),
            indent=2,
        )
    )


@group.command(
    "query-layers", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def query_layers(options, query):
    """
    Query across all layers
    """
    click.echo(
        json.dumps(query_info_output(options, "LayerData[]", None, query), indent=2)
    )



from hamlet.backend.engine.common import ENGINE_GLOBAL_NAME


def run(engine, clear_engine):
    clear_engine()

    global_engine = engine.engine_store.get_engine(ENGINE_GLOBAL_NAME)
    global_engine.install()

    unicycle_engine = engine.engine_store.find_engine("unicycle")
    unicycle_engine.install()

    engine.engine_store.global_engine = "unicycle"

    assert global_engine.installed is True
    assert unicycle_engine.installed is True

    assert engine.engine_store.global_engine == "unicycle"



import click
from hamlet.backend.run import sentry_release as run_sentry_release_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@click.command(
    "sentry-release",
    short_help="Upload sourcemap files to sentry for a specific release",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-m", "--sentry-source-map-s3-url", required=True, help="s3 link to sourcemap files"
)
@click.option("-p", "--sentry-url-prefix", help="prefix for sourcemap files")
@click.option("-r", "--sentry-release-name", help="sentry release name", required=True)
@click.option(
    "-s", "--run-setup", help="run setup installation to prepare", is_flag=True
)
@exceptions.backend_handler()
@pass_options
def sentry_release(options, **kwargs):
    """
    Upload sourcemap files to sentry for a specific release
    """

    args = {**options.opts, **kwargs}

    run_sentry_release_backend.run(**args, _is_cli=True)



import os
import click
import json

from tabulate import tabulate

from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.backend import query as query_backend


def query_info_output(options, query, query_params=None, sub_query_text=None):
    query_args = {
        **options.opts,
        "generation_entrance": "info",
        "output_filename": "info.json",
        "use_cache": False,
    }
    query_result = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        query_text=query,
        query_params=query_params,
        sub_query_text=sub_query_text
    )

    return query_result


def component_types_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Type"]),
                wrap_text(row["Description"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Type", "Description"],
        tablefmt="github",
    )


@click.command(
    "list-component-types", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="A query to filter out the results")
@json_or_table_option(component_types_table)
@exceptions.backend_handler()
@pass_options
def list_component_types(options, query):
    """
    Lists the types of components available
    """

    LIST_COMPONENT_TYPES_QUERY = (
        "ComponentTypes[]" ".{" "Type:Type," "Description:Description" "}"
    )

    return query_info_output(options, LIST_COMPONENT_TYPES_QUERY, None, query)


@click.command(
    "describe-component-type",
    short_help="",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-t", "--type", required=True, help="The type of the component to describe"
)
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def describe_component_type(options, type, query):
    """
    Describes a specific component type
    """
    DESCRIBE_COMPONENT_TYPE_QUERY = "ComponentTypes[?Type=={type}] | [0]"

    click.echo(
        json.dumps(
            query_info_output(
                options, DESCRIBE_COMPONENT_TYPE_QUERY, {"type": type}, query
            ),
            indent=2,
        )
    )


@click.command(
    "query-component-types", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="a query on the describe result")
@exceptions.backend_handler()
@pass_options
def query_component_types(options, query):
    """
    Query across all component types
    """
    click.echo(
        json.dumps(
            query_info_output(options, "ComponentTypes[]", None, query), indent=2
        )
    )



from hamlet.command import root as cli
from .stack import stack as stack_cmd
from .deployment import deployment as deployment_cmd
from .crypto import crypto as crypto_cmd
from .file_crypto import file_crypto as file_crypto_cmd
from .credentials_crypto import credentials_crypto as credentials_crypto_cmd


@cli.group("manage", context_settings=dict(max_content_width=240))
def group():
    """
    Manages stuff
    """


group.add_command(deployment_cmd)
group.add_command(stack_cmd)
group.add_command(crypto_cmd)
group.add_command(file_crypto_cmd)
group.add_command(credentials_crypto_cmd)



import os
import click
from hamlet.backend.test.generate import run as test_generate_backend
from hamlet.command.common import exceptions


@click.command(
    "generate",
    context_settings=dict(max_content_width=240),
    short_help="Generate tests based on .tescase.json files",
)
@click.option(
    "-f",
    "--filename",
    "filenames",
    multiple=True,
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="path to testcase file",
)
@click.option(
    "-d",
    "--directory",
    "directory",
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    default=".",
    show_default=True,
    help="directory to scan for testcase files[depth=1]",
)
@click.option(
    "-o",
    "--output",
    "output",
    type=click.Path(file_okay=True, dir_okay=False),
    help="output file path",
)
@exceptions.backend_handler()
def geneate(filenames, output, directory):
    """
    Used to generate pytests from testcase files.

    \b
    Test type determined by target file extension:

    \b
        1. ".json" - cf template test

    \b
    Note:
    \b
        1. When no files and no directory provided current directory is used to scan for testcase files(depth=1).
        2. Multiple testcase files merged in runtime therefore all testcases must have unique names.
        3. When no output file path provided output will be sent to stdout
        4. Tescase files must have ".testcase.json" extension. Otherwise command will complain and not run.

    \b
    Testcase file structure:

    \b
    {
        "casename":{
            "filename": "/path/to/subject.json",
            "no_lint":True,
            "no_vulnerability_check":True,
            "structure": {
                "exists":[
                    "path.in.json.obj",
                    "path.in.json.array[0].test"
                ],
                "match":[
                    [
                        "path.to.json.object",
                        {
                            "key": "value"
                        }
                    ],
                    [
                        "path.to.json.scalar",
                        "scalar value"
                    ],
                    [
                        "path.to.json.value.in.array[10]",
                        -5
                    ]
                ],
                "resource":[
                    [
                        "resourceName",
                        "resourceType"
                    ]
                ],
                "output":[
                    "outputName"
                ],
                "not_empty":[
                    "not.empty.path"
                ],
                "length":[
                    "path.to.json.array",
                    10
                ]
            }
        }
    }

    """
    directory = os.path.abspath(directory) if directory == "." else directory
    result = test_generate_backend(
        filenames=filenames, output=output, directory=directory
    )
    if not output:
        click.echo(result)



from hamlet.backend.common.exceptions import BackendException


class HamletEngineInvalidVersion(BackendException):
    """
    Raise when the engine found in state does not align with the version the cli supports
    """

    def __init__(self, engine_name, version):
        self.version = version
        self.engine_name = engine_name

        msg = (
            f"Engine {self.engine_name} state not supported by this cli version"
            f"- run: hamlet engine install-engine {self.engine_name}"
        )
        super().__init__(msg)


class EngineStoreException(BackendException):
    """
    A base exception for issues in the Engine Store
    """

    pass


class EngineStoreMissingEngineException(EngineStoreException):
    """
    A requested engine can't be found in the engine store
    """

    pass



import functools
import click
from click._compat import get_text_stderr

import httpx._exceptions as httpx_exceptions
from click.exceptions import ClickException

from hamlet.backend.common.exceptions import BackendException
from hamlet.env import HAMLET_GLOBAL_CONFIG


class CommandError(ClickException):
    """An exception occurred during processing"""

    def show(self, file=None):
        if file is None:
            file = get_text_stderr()
        click.echo(
            click.style("[!] Error Running Command", fg="red", bold=True), err=True
        )
        click.echo("")
        click.echo(self.format_message(), file=file, err=True)

    def format_message(self):
        return click.style(str(self.message))


class HamletHomeDirUnavailableException(CommandError):
    def show(self):
        click.secho(
            (
                f"[!] The hamlet home dir {HAMLET_GLOBAL_CONFIG.home_dir} is not assessible\n"
                f"[!] When testing access the reported error was:\n\n"
                f"  {self.message} \n\n"
                "[!] Check the permissions on the directory"
                " or change your home dir using the HAMLET_ENGINE_DIR environment variable\n"
            ),
            fg="red",
            bold=True,
            err=True,
        )

    def format_message(self):
        return click.style(str(self.message))


def backend_handler():
    """Handles backend errors with a formatted click exception message"""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)

            except httpx_exceptions.RequestError as e:

                msg = f"HTTP error | url: {e.request.url} | msg: {str(e)}"
                raise CommandError(msg)

            except BackendException as e:

                raise CommandError(str(e))

        return wrapper

    return decorator



import click
from hamlet.backend.manage import stack as manage_stack_backend
from hamlet.command.common.config import pass_options
from hamlet.command.common import exceptions


@click.command(
    "stack",
    short_help="Manage a CloudFormation stack",
    context_settings=dict(max_content_width=240),
)
@click.option("-d", "--delete", help="delete the stack", is_flag=True)
@click.option(
    "-i",
    "--stack-initiate",
    is_flag=True,
    help="initiate but do not monitor the stack operation(disable monitoring)",
)
@click.option(
    "-m",
    "--stack-monitor",
    is_flag=True,
    help="monitor but do not initiate the stack operation(disable initiation)",
)
@click.option(
    "-w",
    "--stack-wait",
    type=click.INT,
    show_default=True,
    default=30,
    help="interval between checking the progress of the stack operation",
)
@click.option(
    "-r",
    "--region",
    help="AWS region identifier for the region in which the stack should be managed",
)
@click.option(
    "-l", "--deployment-group", help="the deployment group to manage", required=True
)
@click.option(
    "-u",
    "--deployment-unit",
    help="deployment unit used to determine the stack template",
    required=True,
)
@click.option(
    "-z", "--deployment-unit-subset", help="subset of the deployment unit required"
)
@click.option(
    "-y",
    "--dryrun",
    is_flag=True,
    help="show what will happen without actually updating the stack",
)
@exceptions.backend_handler()
@pass_options
def stack(options, **kwargs):
    """
    Manage a CloudFormation stack

    \b
    NOTES:
    1. You must be in the correct directory corresponding to the requested stack level
    2. REGION is only relevant for the "product" level, where multiple product stacks are necessary
       if the product uses resources in multiple regions
    4. If stack doesn't exist in AWS, the update operation will create the stack
    5. Overriding the stack name is not recommended except where legacy naming has to be maintained
    6. A dryrun creates a change set, then displays it. It only applies when
       the STACK_OPERATION=update
    """

    args = {**options.opts, **kwargs}

    manage_stack_backend.run(**args, _is_cli=True)



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.manage.file_crypto import file_crypto as manage_file_crypto
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["-d,--decrypt"] = [True, False]
ALL_VALID_OPTIONS["-e,--encrypt"] = [True, False]
ALL_VALID_OPTIONS["-f,--crypto-file"] = "crypto_file"
ALL_VALID_OPTIONS["-u,--update"] = [True, False]


@mock.patch("hamlet.command.manage.file_crypto.manage_file_crypto_backend")
def test_input_valid(manage_file_crypto_backend):
    runner = CliRunner()
    with runner.isolated_filesystem():
        open("crypto_file", "w").close()
        run_options_test(
            runner,
            manage_file_crypto,
            ALL_VALID_OPTIONS,
            manage_file_crypto_backend.run,
        )


@mock.patch("hamlet.command.manage.file_crypto.manage_file_crypto_backend")
def test_input_validation(manage_file_crypto_backend):
    runner = CliRunner()
    with runner.isolated_filesystem():
        open("crypto_file", "w").close()
        run_validatable_option_test(
            runner,
            manage_file_crypto,
            manage_file_crypto_backend.run,
            {},
            [("-f", "nonexistingfile", "crypto_file")],
        )



from hamlet.backend.automation_tasks.base import AutomationRunner
from hamlet.backend.automation import (
    properties_file,
    set_automation_context,
    construct_tree,
    confirm_builds,
    update_build_references,
)


class UpdateBuildAutomationRunner(AutomationRunner):
    def __init__(
        self,
        deployment_unit=None,
        build_reference=None,
        code_tag=None,
        image_format=None,
        registry_scope=None,
        **kwargs
    ):
        super().__init__()

        self._context_env = {
            "DEPLOYMENT_UNITS": deployment_unit,
            "GIT_COMMIT": build_reference,
            "CODE_TAGS": code_tag,
            "IMAGE_FORMATS": image_format,
            "REGISTRY_SCOPE": registry_scope,
            "DEFER_REPO_PUSH": "true",
            **kwargs,
        }

        self._script_list = [
            {
                "func": construct_tree.run,
                "args": {
                    "use_existing_tree": True,
                    "_is_cli": True,
                },
            },
            {
                "func": properties_file.get_automation_properties,
                "args": {**self._context_env},
            },
            {"func": set_automation_context.run, "args": {"_is_cli": True}},
            {"func": confirm_builds.run, "args": {"_is_cli": True}},
            {"func": update_build_references.run, "args": {"_is_cli": True}},
        ]



import subprocess

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop

with open("README.md", "r") as fh:
    long_description = fh.read()

about = {}
with open("hamlet/__about__.py", "r") as fp:
    exec(fp.read(), about)

packages = find_packages(exclude=["tests.*", "tests"])


class InstallCommand(install):
    def __post_install(self, dir):
        subprocess.call(["./setup-hamlet.sh"])

    def run(self):
        install.run(self)
        self.execute(
            self.__post_install, (self.install_lib,), msg="Setting up hamlet..."
        )


class DevelopCommand(develop):
    def __post_install(self, dir):
        subprocess.call(["./setup-hamlet.sh"])

    def run(self):
        develop.run(self)
        self.execute(
            self.__post_install, (self.install_lib,), msg="Setting up hamlet..."
        )


setup(
    name=about["__title__"],
    author=about["__author__"],
    url=about["__url__"],
    description=about["__summary__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv3",
    project_urls={"Repository": about["__repository_url__"]},
    packages=packages,
    setup_requires=["setuptools_scm>=6.0.1,<7.0.0"],
    use_scm_version={
        "root": "",
        "relative_to": __file__,
        "fallback_version": "_testing_",
        "write_to": "hamlet/__version__.py",
        "local_scheme": "no-local-version",
    },
    install_requires=[
        "click>=8.0.3,<9.0.0",
        "click-configfile>=0.2.3,<1.0.0",
        "cookiecutter>=1.7.0,<2.0.0",
        "tabulate>=0.8.0,<1.0.0",
        "Jinja2>=2.11.0,<3.0.0",
        "jmespath>=0.10.0,<1.0.0",
        "importlib-resources>=5.2.0,<6.0.0",
        "www-authenticate>=0.9.2,<1.0.0",
        "httpx>=0.19.0,<0.20.0",
        "marshmallow>=3.7.0,<4.0.0",
        # Template testing
        "pytest>=6.0.0,<7.0.0",
        # AWS testing dependencies
        "cfn-lint>=0.54.1,<1.0.0",
        # - boto3 locked to support checkov current requirements
        "checkov>=2.0.461,<3.0.0",
        "boto3>=1.17.111,<1.18.0",
        "botocore>=1.20.112,<2.0.0",
        # Diagrams
        "diagrams>=0.18.0,<1.0.0",
    ],
    include_package_data=True,
    python_requires=">=3.6",
    entry_points={"console_scripts": ["hamlet=hamlet:command.root"]},
    cmdclass={
        "install": InstallCommand,
        "develop": DevelopCommand,
    },
    classifiers=[
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Java",
        "Programming Language :: Python",
    ],
)



from hamlet.command import root as cli

from .create import create_deployments as create_deployments_cmd
from .list import list_deployments as list_deployments_cmd
from .run import run_deployments as run_deployments_cmd
from .test import test_deployments as test_deployments_cmd


@cli.group("deploy", context_settings=dict(max_content_width=240))
def group():
    """
    Deploys infrastructure based on the hamlet cmdb
    """
    pass


group.add_command(create_deployments_cmd)
group.add_command(list_deployments_cmd)
group.add_command(run_deployments_cmd)
group.add_command(test_deployments_cmd)



import os

from hamlet.env import HAMLET_GLOBAL_CONFIG

ENGINE_STORE_DEFAULT_DIR = os.path.join(HAMLET_GLOBAL_CONFIG.home_dir, "engine")
ENGINE_GLOBAL_NAME = "_global"
ENGINE_DEFAULT_GLOBAL_ENGINE = "train"
ENGINE_STATE_FILE_NAME = "engine_state.json"

ENGINE_STATE_VERSION = "1.0.1"



import os
import tempfile
from pytest import raises
from hamlet.backend.test.run import run as run_test_backend
from hamlet.backend.test.generate import run as generate_test_backend
from .conftest import DATA_DIR


def test():
    with tempfile.TemporaryDirectory() as dir:
        output_filename = os.path.join(dir, "test.py")

        def testcase_filename(name):
            return os.path.join(DATA_DIR, "testcase", name)

        generate_test_backend(
            [testcase_filename("secure-testcase.json")], output_filename
        )
        assert run_test_backend([output_filename])
        generate_test_backend(
            [testcase_filename("insecure-testcase.json")], output_filename
        )
        with raises(Exception) as e:
            run_test_backend([output_filename])
            assert e.message == "Tests Failed"

        generate_test_backend(
            [testcase_filename("valid-syntax-testcase.json")], output_filename
        )
        assert run_test_backend([output_filename])
        generate_test_backend(
            [testcase_filename("invalid-syntax-testcase.json")], output_filename
        )
        with raises(Exception) as e:
            run_test_backend([output_filename])
            assert e.message == "Tests Failed"

        generate_test_backend(
            [testcase_filename("structure-good-testcase.json")], output_filename
        )
        assert run_test_backend([output_filename])
        generate_test_backend(
            [testcase_filename("structure-bad-testcase.json")], output_filename
        )
        with raises(Exception) as e:
            run_test_backend([output_filename])
            assert e.message == "Tests Failed"

        generate_test_backend(
            [testcase_filename("json-structure-good-testcase.json")], output_filename
        )
        assert run_test_backend([output_filename])
        generate_test_backend(
            [testcase_filename("json-structure-bad-testcase.json")], output_filename
        )
        with raises(Exception) as e:
            run_test_backend([output_filename])
            assert e.message == "Tests Failed"



import typing

from hamlet.backend.engine.engine_loader import EngineLoader

from hamlet.backend.engine.engine import Engine
from hamlet.backend.engine.engine_part import (
    CoreEnginePart,
    AWSEnginePluginPart,
    AzureEnginePluginPart,
    BashExecutorEnginePart,
    WrapperEnginePart,
)
from hamlet.backend.engine.engine_source import ContainerEngineSource


class UnicycleEngineLoader(EngineLoader):
    """
    Provides the latest builds of all official hamlet components
    Each component is sourced directly from the image that is created on commit to the default branch
    """

    def load(self) -> typing.Iterable[Engine]:
        engine_sources = [
            ContainerEngineSource(
                name="engine",
                description="hamlet core engine",
                registry_url="https://ghcr.io",
                repository="hamlet-io/engine",
                tag="edge",
            ),
            ContainerEngineSource(
                name="executor-bash",
                description="hamlet bash executor",
                registry_url="https://ghcr.io",
                repository="hamlet-io/executor-bash",
                tag="edge",
            ),
            ContainerEngineSource(
                name="engine-plugin-aws",
                description="hamlet aws engine plugin",
                registry_url="https://ghcr.io",
                repository="hamlet-io/engine-plugin-aws",
                tag="edge",
            ),
            ContainerEngineSource(
                name="engine-plugin-azure",
                description="hamlet azure engine plugin",
                registry_url="https://ghcr.io",
                repository="hamlet-io/engine-plugin-azure",
                tag="edge",
            ),
            ContainerEngineSource(
                name="engine-core",
                description="hamlet freemarker wrapper",
                registry_url="https://ghcr.io",
                repository="hamlet-io/engine-core",
                tag="edge",
            ),
        ]

        engine_parts = [
            WrapperEnginePart(source_path="", source_name="engine-core"),
            CoreEnginePart(source_path="", source_name="engine"),
            BashExecutorEnginePart(source_path="", source_name="executor-bash"),
            AWSEnginePluginPart(source_path="", source_name="engine-plugin-aws"),
            AzureEnginePluginPart(source_path="", source_name="engine-plugin-azure"),
        ]

        engine = Engine(
            name="unicycle", description="Latest build of the official engine parts"
        )
        engine.parts = engine_parts
        engine.sources = engine_sources

        yield engine



import os.path


__all__ = [
    "__title__",
    "__summary__",
    "__url__",
    "__repository_url__",
    "__author__",
    "__email__",
]


try:
    base_dir = os.path.dirname(os.path.abspath(__file__))
except NameError:
    base_dir = None

__title__ = "hamlet"
__summary__ = "Building your infrastructure"
__url__ = "https://hamlet.io"
__repository_url__ = "https://github.com/hamlet-io/executor-python"

__author__ = "Hamlet"
__email__ = "help@hamlet.io"



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.task import task as run_task
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-i,--component"] = "container_name"
ALL_VALID_OPTIONS["!-w,--task"] = "task_name"
ALL_VALID_OPTIONS["!-t,--tier"] = "tier"
ALL_VALID_OPTIONS["-c,--container-id"] = "container_id"
ALL_VALID_OPTIONS["-d,--delay"] = 100
ALL_VALID_OPTIONS["-e,--env"] = "somevar"
ALL_VALID_OPTIONS["-j,--component-instance"] = "component_instance"
ALL_VALID_OPTIONS["-k,--component-version"] = "component_version"
ALL_VALID_OPTIONS["-v,--value"] = "value"
ALL_VALID_OPTIONS["-x,--instance"] = "task_instance"
ALL_VALID_OPTIONS["-y,--version"] = "task_version"


@mock.patch("hamlet.command.run.task.run_task_backend")
def test_input_valid(run_task_backend):
    run_options_test(CliRunner(), run_task, ALL_VALID_OPTIONS, run_task_backend.run)


@mock.patch("hamlet.command.run.task.run_task_backend")
def test_input_validation(run_task_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        run_task,
        run_task_backend.run,
        {
            "-i": "container_name",
            "-w": "task_name",
            "-t": "tier",
        },
        [("-d", "not_an_int", 10)],
    )



from hamlet.command import root as cli
from .cmdb import generate_tenant as generate_tenant_cmdb_cmd
from .cmdb import generate_account as generate_account_cmdb_cmd
from .cmdb import generate_product as generate_product_cmdb_cmd


@cli.group("generate", context_settings=dict(max_content_width=240))
def group():
    """
    Generates base CMDB file system structures
    """
    pass


group.add_command(generate_tenant_cmdb_cmd)
group.add_command(generate_account_cmdb_cmd)
group.add_command(generate_product_cmdb_cmd)



"""
This module's idea is the automation of very repetitive tests which we must have
to test the correctness of our cli configuration. This module's functions greatly reduce the number of lines
that we need to maintain. And because of that tests become much more accurate
and reliable.

Sure, manual individual testing is somewhat better but I think that we'll test "backend" part of our script separately.
Therefore this module will be usefull even after we'll rewrite everything in python.
"""


import copy
import collections
from unittest import mock
import pytest
import click
from click.testing import CliRunner
from hamlet.loggers import logging


logger = logging.getLogger("options")


def are_required_keys(keys):
    return keys.startswith("!")


def parse_keys(keys):
    # removing spaces and required flag
    return tuple(key.strip().strip("!") for key in keys.split(",") if key)


def get_key(keys, index):
    keys = parse_keys(keys)
    try:
        return keys[index]
    except IndexError:
        return keys[-1]


def is_values_collection(values):
    return isinstance(values, collections.abc.Iterable) and not isinstance(values, str)


def generate_incremental_required_options_collection(all_options):
    """
    Generates collection of cli required option(marked by!)
    Used to test that command can't run without all being given all required options.
    For example we have 3 required options with an alternative key each.
    Function will produce generator which will give:
    ([], True),
    (
        [
            -first-required-option, value
        ],
        True
    ),
    (
        [
            -first-required-option, value,
            -second-required-option, value
        ],
        True
    ),
    (
        [
            -first-required-option, value,
            -second-required-option, value
            -third-required-option, value
        ],
        False
    )
    Where first element of tuple is args list and second element is bool error flag.
    Error flag used to select correct check for given args list.
    """
    required_options = collections.OrderedDict()
    required_keys_max_parts = 1
    for keys, values in all_options.items():
        if are_required_keys(keys):
            if is_values_collection(values):
                value = values[0]
            else:
                value = values
            required_options[keys] = value
            required_keys_max_parts = max(
                required_keys_max_parts, len(parse_keys(keys))
            )

    if not required_options:
        logger.info(["no required options"])
        return

    for key_index in range(required_keys_max_parts):
        for max_arguments in range(len(required_options) + 1):
            args = []
            str_args = []
            for keys, value in required_options.items():
                if len(str_args) >= max_arguments:
                    break
                key = get_key(keys, key_index)
                args.append(key)
                args.append(value)
                str_args.append("{} {}".format(key, value))
            logger.info(
                "[required options: %s/%s]", len(str_args), len(required_options)
            )
            logger.info("\n".join(str_args))
            yield args, len(str_args) < len(required_options)


def generate_test_options_collection(all_options):
    """
    Generates collection of CLI options based on OrderedDict with next structure:
        ordered_dict['-o, --option'] = 'value'
        ordered_dict['-m', '--multi'] = ['a', 'b']
    Results:
        [
            '-o', 'value',
            '-m', 'a'
        ],
        [
            '--option', 'value',
            '--multi', 'a'
        ],
        [
            '-o', 'value',
            '-m', 'b'
        ],
        [
            '--option', 'value',
            '--multi', 'b'
        ]
    Function does not create all posible combinations. Instead it freezes first value of all keys with
    iterable collection values and then creates variations by iterating over one of the keys with collection values
    untill all of them iterated.
    """
    required_options = collections.OrderedDict()
    all_keys_max_parts = 1
    required_keys_max_parts = 1
    required_keys_collection_values = []
    optional_keys_collection_values = []

    for keys, values in all_options.items():
        key_parts = parse_keys(keys)
        all_keys_max_parts = max(all_keys_max_parts, len(key_parts))
        required = are_required_keys(keys)
        if required:
            required_options[keys] = values
            required_keys_max_parts = max(required_keys_max_parts, len(key_parts))
        if is_values_collection(values):
            if required:
                required_keys_collection_values.append(keys)
            else:
                optional_keys_collection_values.append(keys)

    # fixed_value_keys is plural because it's a commaseparated list of keys
    def generate_args(
        key_index, fixed_value_keys=None, fixed_value_index=0, options=None
    ):
        args = []
        str_args = []
        for keys, values in options.items():
            key = get_key(keys, key_index)
            if is_values_collection(values):
                if keys == fixed_value_keys:
                    value = values[fixed_value_index]
                else:
                    value = values[0]
            else:
                value = values
            if isinstance(value, bool):
                if value:
                    args.append(key)
                    str_args.append(key)
            else:
                args.append(key)
                args.append(value)
                str_args.append("{} {}".format(key, value))
        if fixed_value_index == 0:
            logger.info("\n".join(str_args))
        else:
            logger.info(
                "Change:[%s %s]",
                get_key(fixed_value_keys, key_index),
                options[fixed_value_keys][fixed_value_index],
            )
        return args

    def generate_args_group(max_keys_parts, options, collection_values_keys):
        for key_index in range(max_keys_parts):
            yield generate_args(key_index, options=options)
        # starting from index 1 because previous generation conains all 0 indexes
        if collection_values_keys:
            for fixed_value_keys in collection_values_keys:
                for fixed_value_index in range(1, len(options[fixed_value_keys])):
                    for key_index in range(max_keys_parts):
                        yield generate_args(
                            key_index, fixed_value_keys, fixed_value_index, options
                        )

    # generating using only required options
    if len(required_options) != len(all_options):
        logger.info("[required]")
        if required_options:
            for args in generate_args_group(
                required_keys_max_parts,
                required_options,
                required_keys_collection_values,
            ):
                yield args
        else:
            # command should run without options
            logger.info("None")
            yield []

    # generating using all options
    logger.info("[all]")
    for args in generate_args_group(
        all_keys_max_parts, all_options, optional_keys_collection_values
    ):
        yield args


# pure utility function
def options_dict_to_list(options):
    args = []
    for key, value in options.items():
        args.append(key)
        args.append(value)
    return args


def run_single_validatable_option_test(
    runner,
    cmd,
    runner_mock,
    options,
    key,
    incorrect,
    correct,
):
    """
    Test option validation rule by providing incorrect and correct values for the same option.
    That's is not the best test in the world, but click env is quite isolated.
    Therefore I can't capture click's standard exceptions.
    """
    logger.info("[validation test:%s]", key)
    logger.info("invalid value:%s", incorrect)
    options = copy.deepcopy(options)
    options[key] = incorrect
    result = runner.invoke(cmd, options_dict_to_list(options))
    assert result.exit_code == 2, result.output
    assert runner_mock.call_count == 0
    logger.info("valid value:%s", correct)
    options[key] = correct
    result = runner.invoke(cmd, options_dict_to_list(options))
    assert result.exit_code == 0, result.output
    assert runner_mock.call_count == 1
    runner_mock.call_count = 0
    logger.info("[success]")


def run_validatable_option_test(runner, cmd, runner_mock, options, tests):
    """
    Handy shortcut for testing multiply options of the same command using run_single_validatable_option_test
    """
    for key, incorrect, correct in tests:
        run_single_validatable_option_test(
            runner, cmd, runner_mock, options, key, incorrect, correct
        )


def run_options_test(runner, cmd, options, runner_mock):
    """
    Shortcut which runs generalized tests:
    1. Testing that no more options exist
    2. Testing that required options are required.
    3. Testing that all options accepted and all expected values pass validation.
    """
    assert len(cmd.params) == len(options)
    for args, error in generate_incremental_required_options_collection(options):
        result = runner.invoke(cmd, args)
        if error:
            assert result.exit_code == 2, result.output
            assert runner_mock.call_count == 0
        else:
            logger.info(result.exception)
            assert result.exit_code == 0, result.output
            assert runner_mock.call_count == 1
            runner_mock.call_count = 0

    for args in generate_test_options_collection(options):
        result = runner.invoke(cmd, args)
        logger.info(result.exc_info)
        assert result.exit_code == 0, result.output
        assert runner_mock.call_count == 1
        runner_mock.call_count = 0


def test_run_options_test():

    runner_mock = mock.MagicMock()
    runner = CliRunner()

    @click.command()
    @click.option("-a", "--a", required=True)
    @click.option("-b", "--b", required=True, type=click.Choice(["bvalue1", "bvalue2"]))
    @click.option("-c", "--c", type=click.Choice(["cvalue1", "cvalue2", "cvalue3"]))
    @click.option("-f", "--f", is_flag=True)
    def test_command(a, b, c, f):
        runner_mock()

    def run_test(options, error=True):
        if error:
            with pytest.raises(AssertionError):
                run_options_test(runner, test_command, options, runner_mock)
        else:
            run_options_test(runner, test_command, options, runner_mock)

    correct_options = collections.OrderedDict()
    correct_options["!-a,--a"] = "avalue"
    correct_options["!-b,--b"] = ["bvalue1", "bvalue2"]
    correct_options["-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    correct_options["-f,--f"] = [True, False]

    # tests should pass for correct options
    run_test(correct_options, False)

    not_all_options = collections.OrderedDict()
    not_all_options["!-a,--a"] = "avalue"
    not_all_options["!-b,--b"] = ["bvalue1", "bvalue2"]
    not_all_options["-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]

    # should fail because not all options listed
    run_test(not_all_options)

    incorrect_value_options = collections.OrderedDict()
    incorrect_value_options["!-a,--a"] = "avalue"
    incorrect_value_options["!-b,--b"] = ["bvalue1", "bvalue2", "bvalue3"]
    incorrect_value_options["-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    incorrect_value_options["-f,--f"] = [True, False]

    # should fail because bvalue3 is not accepted
    run_test(incorrect_value_options)

    required_is_optional_options = collections.OrderedDict()
    required_is_optional_options["!-a,--a"] = "avalue"
    required_is_optional_options["-b,--b"] = ["bvalue1", "bvalue2"]
    required_is_optional_options["-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    required_is_optional_options["-f,--f"] = [True, False]

    # should fail because b is not required
    run_test(required_is_optional_options)

    optional_is_required_options = collections.OrderedDict()
    optional_is_required_options["!-a,--a"] = "avalue"
    optional_is_required_options["!-b,--b"] = ["bvalue1", "bvalue2"]
    optional_is_required_options["!-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    optional_is_required_options["-f,--f"] = [True, False]

    # should fail because c is required
    run_test(optional_is_required_options)


def test_validatable_options_test():

    runner_mock = mock.MagicMock()
    runner = CliRunner()

    @click.command()
    @click.option("-r", required=True, type=click.INT)
    @click.option("-o", type=click.INT)
    def test_command(r, o):
        runner_mock()

    def run_test(options, error=True):
        def start():
            run_validatable_option_test(
                runner, test_command, runner_mock, {"-r": "10"}, options
            )

        if error:
            with pytest.raises(AssertionError):
                start()
        else:
            start()

        # will raise error because all values valid
        run_test([("-r", "10", "10"), ("-o", "not_an_int", "10")])
        # will raise error because -r option incorrect value passes validation
        run_test([("-r", "10", "not_an_int"), ("-o", "not_an_int", "10")])
        # will raise error because -o option all invalid
        run_test([("-r", "not_an_int", "10"), ("-o", "not_an_int", "not_an_int")])
        # tests must pass
        run_test([("-r", "not_an_int", "10"), ("-o", "not_an_int", "10")], False)


def test_generate_incremental_required_options_collection():
    options = collections.OrderedDict()
    options["!-a,--a"] = "avalue"
    options["-b,--b"] = ["bvalue1", "bvalue2"]
    options["!-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    options["-f,--f"] = [True, False]

    generator = generate_incremental_required_options_collection(options)
    assert next(generator) == ([], True)
    assert next(generator) == (["-a", "avalue"], True)
    assert next(generator) == (["-a", "avalue", "-c", "cvalue1"], False)
    assert next(generator) == ([], True)
    assert next(generator) == (["--a", "avalue"], True)
    assert next(generator) == (["--a", "avalue", "--c", "cvalue1"], False)
    with pytest.raises(StopIteration):
        next(generator)


def test_option_generation():
    options = collections.OrderedDict()
    options["!-a,--a"] = "avalue"
    options["-b,--b"] = ["bvalue1", "bvalue2"]
    options["!-c,--c"] = ["cvalue1", "cvalue2", "cvalue3"]
    options["-f,--f"] = [True, False]
    generator = generate_test_options_collection(options)
    # start with  only required options
    assert next(generator) == ["-a", "avalue", "-c", "cvalue1"]
    assert next(generator) == ["--a", "avalue", "--c", "cvalue1"]
    # iteration over required collection values
    assert next(generator) == ["-a", "avalue", "-c", "cvalue2"]
    assert next(generator) == ["--a", "avalue", "--c", "cvalue2"]
    assert next(generator) == ["-a", "avalue", "-c", "cvalue3"]
    assert next(generator) == ["--a", "avalue", "--c", "cvalue3"]
    # then all options without required collection values
    assert next(generator) == ["-a", "avalue", "-b", "bvalue1", "-c", "cvalue1", "-f"]
    assert next(generator) == [
        "--a",
        "avalue",
        "--b",
        "bvalue1",
        "--c",
        "cvalue1",
        "--f",
    ]
    # collection values variations
    # iterating over '-b,--b' collection
    assert next(generator) == ["-a", "avalue", "-b", "bvalue2", "-c", "cvalue1", "-f"]
    assert next(generator) == [
        "--a",
        "avalue",
        "--b",
        "bvalue2",
        "--c",
        "cvalue1",
        "--f",
    ]
    # iterating over '-f,--f'
    assert next(generator) == [
        "-a",
        "avalue",
        "-b",
        "bvalue1",
        "-c",
        "cvalue1",
    ]
    assert next(generator) == [
        "--a",
        "avalue",
        "--b",
        "bvalue1",
        "--c",
        "cvalue1",
    ]
    with pytest.raises(StopIteration):
        next(generator)



from unittest import mock

from hamlet.backend.automation_tasks.base import AutomationRunner


def test_automation_runner_task():
    """
    Tests that a task in the runner has been called
    """

    mock_task = mock.Mock()

    automation_runner = AutomationRunner()
    automation_runner._script_list = [
        {"func": mock_task, "args": {"mock_task_arg": "mock_task_arg_value"}}
    ]

    automation_runner.run()

    mock_task.assert_called_once_with(
        env=automation_runner._context_env, mock_task_arg="mock_task_arg_value"
    )

    assert "AUTOMATION_PROVIDER" in automation_runner._context_env
    assert "AUTOMATION_DATA_DIR" in automation_runner._context_env



import click
import functools

from hamlet.command import root as cli
from hamlet.command.common import exceptions, config

from hamlet.backend.automation_tasks import upload_image as upload_image_backend
from hamlet.backend.automation_tasks import transfer_image as transfer_image_backend
from hamlet.backend.automation_tasks import update_build as update_build_backend


def image_options(func):
    """Standard image details"""

    @click.option(
        "-u",
        "--deployment-unit",
        required=True,
        help="The deployment unit the image belongs to",
    )
    @click.option(
        "-r",
        "--build-reference",
        required=True,
        help="The unique reference for the build of this image - usually git commit",
    )
    @click.option(
        "--code-tag",
        help="A tag applied to your code repo which will be mapped to a build reference",
    )
    @click.option(
        "-f",
        "--image-format",
        help="The format of the code image",
        type=click.Choice(
            [
                "dataset",
                "rdssnapshot",
                "docker",
                "lambda",
                "pipeline",
                "scripts",
                "openapi",
                "swagger",
                "spa",
                "contentnode",
            ]
        ),
    )
    @click.option(
        "-s",
        "--registry-scope",
        help="The scope of the registry to update the image to",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """
        Image Detail handling
        """
        return func(*args, **kwargs)

    return wrapper


@cli.group("release", context_settings=dict(max_content_width=240))
def group():
    """
    Manage the lifecycle of code images
    """


@group.command(
    "upload-image", short_help="", context_settings=dict(max_content_width=240)
)
@image_options
@click.option(
    "--image-path",
    help="The path to the image can be zip file or directory",
    type=click.Path(file_okay=True, dir_okay=True, readable=True, resolve_path=True),
)
@click.option(
    "--dockerfile",
    help="The path to a dockerfile to create the image with",
    type=click.Path(file_okay=True, dir_okay=False, readable=True, resolve_path=True),
)
@click.option(
    "--docker-context",
    help="The docker context directory used with the dockerfile",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
)
@click.option("--docker-image", help="The tag of an existing docker image")
@exceptions.backend_handler()
@config.pass_options
def upload_image(opts, **kwargs):
    """
    Upload a code image to a registry
    """
    task = upload_image_backend.UploadImageAutomationRunner(**opts.opts, **kwargs)
    task.run()


@group.command(
    "transfer-image", short_help="", context_settings=dict(max_content_width=240)
)
@image_options
@click.option(
    "--source-account",
    required=True,
    help="The name of the account to get the image from",
)
@click.option(
    "--source-environment",
    required=True,
    help="The name of the environment to get the image from",
)
@exceptions.backend_handler()
@config.pass_options
def transfer_image(opts, **kwargs):
    """
    Transfer an image between registries
    """
    task = transfer_image_backend.TransferImageAutomationRunner(**opts.opts, **kwargs)
    task.run()


@group.command(
    "update-image-reference",
    short_help="",
    context_settings=dict(max_content_width=240),
)
@image_options
@exceptions.backend_handler()
@config.pass_options
def update_image_reference(opts, **kwargs):
    """
    Update the image reference for a component
    """
    task = update_build_backend.UpdateBuildAutomationRunner(**opts.opts, **kwargs)
    task.run()
